namespace eval HAMEN {
   namespace export *

   # --------------------------------------------------------------------
   # initialize parameters for HAMEN
   # --------------------------------------------------------------------
   proc init {} {
      global gbl sys wk

      # register to menu bar, and add key binding
      register_backend "Wavefront Method" hamen w "HAMEN::parameter_window"

      # declare constants
      set sys(hamen,COMMAND) "$sys(HAMEN_DIR)/hamen"
      set sys(hamen,CDF_COMMAND) "$sys(HAMEN_DIR)/hamenCDF"
      

      # declare variables
      set gbl(hamen,mformat) 1
      set gbl(hamen,mboundary) 2
      set gbl(hamen,txformat) 2
      set gbl(hamen,rayout) 5
      set gbl(hamen,dx) 0.2
      set gbl(hamen,dsx) 1.0
      set gbl(hamen,first) 0
      set gbl(hamen,velimage) 0
   }

   # --------------------------------------------------------------------
   # Execute raytracing using HAMEN (wavefront method)
   # --------------------------------------------------------------------
   proc raytracing {} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      
      # If there is no velocity model, return.
      if {$gbl(n_interface) < 1} {
	 bar_message "ERROR: No ray86 model defined."
	 return
      }

      # single source (obs) is required.
      set n 0
      for {set i 1} {$i <= $gbl(src,n_source)} {incr i} {
	 if {$gbl(src,$i,onoff)} {
	    incr n
	    set sx $gbl(src,$i,x)
	    set sz $gbl(src,$i,z)
	 }
      }
      if { $n <= 0 } {
	 tk_messageBox -icon error -type ok\
	     -message "No source is specified."
	 return
      }
      if { $n > 1 } { 
	 tk_messageBox -icon warning -type ok\
	     -message "Plural sources are selected.  The last source parameter is used."
      }

      # file existence check ------------------------------
      if { $gbl(hamen,mformat) == 2 && ![file readable $gbl(file,gmfile)] } {
	 tk_messageBox -icon error -type ok
	 -message "Cannot read grid model file."
	 return
      }
      if { $gbl(hamen,mformat) == 3 && ![file readable $gbl(file,cdffile)] } {
	 tk_messageBox -icon error -type ok\
	     -message "Cannot read netCDF model file."
	 return
      }

      # check external command ------------------------------
      if { ![file executable $sys(hamen,COMMAND)] || 
	   ![file executable $sys(hamen,CDF_COMMAND)] } {
	 tk_messageBox -icon error -type ok\
	     -message "Cannot find external wavefront program."
	 return
      }

      # raycode ------------------------------
      set raycode {}
      # first arrival
      if { $gbl(hamen,first) == 1 } {
	 lappend raycode 0
      }
      # reflection (P-wave)
      for {set i 1} {$i <$gbl(n_interface)} {incr i} {
	 if {$gbl(ray,reflect,$i) == 1} {
	    lappend raycode [expr $i + 1]
	 }
      }
      # converted wave
      #      not yet implemented

      if { [llength $raycode] < 1 } { 
	 tk_messageBox -icon warning -type ok\
	     -message "Specify raytypes before compuatation."
	 return
      }

      # Ray86 file ------------------------------
      set ray86fn [format "./modeling.tmp.%d.ray86" [pid]]
      write_ray86 $ray86fn

      # Config file ------------------------------
      bar_message "Now making config file for hamen..."
      set hamencfg [format "./hamen.tmp.%d.cfg" [pid]]
      set fp [open $hamencfg "w"]
      if { $gbl(hamen,dsx) < 0 } {
	 set dsx $gbl(hamen,dx)
      } else {
	 set dsx $gbl(hamen,dsx)
      }
      puts $fp "\# Hamen config file created by modeling."
      puts $fp [format "model   = %s" $ray86fn]
      puts $fp [format "sx      = %s" $sx]
      puts $fp [format "sz      = %s" $sz]
      puts $fp [format "dx      = %s" $gbl(hamen,dx)]
      puts $fp [format "dsx     = %s" $gbl(hamen,dsx)]
      puts $fp [format "raycode = %s" [join $raycode ","]]
      puts $fp [format "tout    = %s" $gbl(hamen,txformat)]
      puts $fp [format "tfile   = %s" $gbl(file,calculated)]
      puts $fp [format "rout    = %s" $gbl(hamen,rayout)]
      puts $fp [format "rfile   = %s" $gbl(file,rayout)]
      switch -- $gbl(hamen,mformat) {
	 1 {
	    # Ray86 ----------
	    set com $sys(hamen,COMMAND)
	 }
	 2 {
	    # Ray86+gm ----------
	    puts $fp [format "gmodel    = %s" $gbl(file,gmfile)]
	    puts $fp [format "mboundary= %s" $gbl(hamen,mboundary)]
	    set com $sys(hamen,COMMAND)
	 }
	 3 {
	    # netCDF ----------
	    puts $fp [format "cdfmodel = %s" $gbl(file,cdffile)]
	    set com $sys(hamen,CDF_COMMAND)
	 }
      }
      close $fp

      # execute HAMEN ------------------------------
      bar_message "Now raytracing by wavefront method..."
      exec_external_command "$com $hamencfg"

      # plot raypaths
      if { $gbl(hamen,rayout) > 0} { plot_rays }
      # plot TX graph
      TX::plot $gbl(file,calculated)

      # delete temporary file
      file delete -force $ray86fn
      file delete -force $hamencfg

      bar_message "Computation by wavefront method was finished."
   }

   # --------------------------------------------------------------------
   # plotting rays
   # --------------------------------------------------------------------
   proc plot_rays {} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      RAYPATH::clear

      # Check file
      if ![file exists $gbl(file,rayout)] {
	 tk_messageBox -icon error -type ok\
	     -message "Cannot find the raypath file, \"$gbl(file,rayout)\""
	 return
      }

      # color
      RAYPATH::define_color

      # Read and plot ------------------------------
      set type 0
      set fp [open $gbl(file,rayout) "r"]
      set coord {}
      while { [gets $fp line] != -1 } {
	 if { [llength $line] == 0 } {
	    # plot
	    if { [llength $coord] > 1 } {
	       RAYPATH::plot1 $type 1 $coord
	    }
	    set coord {}
	 } else {
	    if { [llength $line] == 2 } {
	       lappend coord $line
	    } else {
	       incr type
	    }
	 }
      }
      if { [llength $coord] > 1 } {
	 RAYPATH::plot1 $type 1 $coord
      }
      # lower raypaths
      RAYPATH::lower
      close $fp
   }
}