# --------------------------------------------------------------------
# open_model ---------------------------------------------------------
#           Read model data from a file (Ray86 or v.in)
# --------------------------------------------------------------------
proc open_model {fmt} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # Does model data already exist?
   if {$gbl(n_interface) > 0} {
      # Save or not?
      set ans [tk_messageBox -title "Save Current Model?"\
		   -message "Do you want to save current model?"\
		   -type yesnocancel -icon question]
      if {$ans=="yes"} {
	 save_model ray86
      } elseif {$ans=="cancel"} {
	 bar_message "Read model was canceled."
	 return
      }
   }
   # select file ----------
   if { $fmt == "ray86"} {
      set types {
	 {"Ray86 files"	{.ray86}	}
	 {"All files"		*}
      }
   } else {
      set types {
	 {"in files"	{.in}	}
	 {"dat files"	{.dat}	}
	 {"All files"		*}
      }
   }
   set fn [tk_getOpenFile -filetypes $types -parent .]
   if { $fn == ""} {
      bar_message "No file is selected."
      return
   }
   # load ----------
   bar_message "Load \"$fn\" ..."
   if { $fmt == "ray86"} {
      read_ray86 $fn
   } else {
      read_vin $fn
   }   
   set wk(undo,n) 0
   # Update ray group parameter
   for {set i 1} {$i <$gbl(n_interface)} {incr i} {
      foreach j {refract  reflect  head} {
	    set gbl(ray,$j,$i) 0
      }
   }
}

# --------------------------------------------------------------------
# read_ray86 ---------------------------------------------------------
#            Read ray86 data file. (Extended Ray86 Version 2.0 or later)
# --------------------------------------------------------------------
# [arguments]
# fn : Ray86 file name.
# --------------------------------------------------------------------
proc read_ray86 {fn} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # Delete old velocity structure --------------------
   clear_all
   
   # Open Ray86 file --------------------
   bar_message "Now reading the model file '$fn' ..."
   
   # file readable check
   if {![file readable $fn] } {
      set fn [lindex [file split $fn] end]
      if [file readable $fn] {
	 tk_messageBox -message "Ray86 file is not found in the specified location.  Use \"$fn\" in the current directory instead." -type ok -icon info
      } else {
	 tk_messageBox -message "Can't find Ray86 file."\
	     -icon warning -title "Can't find Ray86 file"\
	     -type ok
	 set types {
	    {"Ray86 files"	{.ray86}	}
	    {"All files"		*}
	 }
	 set fn [tk_getOpenFile -filetypes $types -parent .]
	 if {$fn == ""} {
	    bar_message "No Ray86 model file is selected."
	    return
	 }
      }
   }
   # check and read Ray86 version ------------------------------
   set wk(read_status) 0
   set version [check_ray86_version $fn]
   switch [expr int( $version )] {
      2 {
	 # version 2 or earlier
	 exec_read_old_ray86 $fn
      }
      3 {
	 # version 3
	 exec_read_ray86 $fn
      }
      -1 {
	 bar_message "Invalid ray86 format."
      }
      default {
	 # unknown
	 bar_message "unknown version of ray86 file."
      }
   }

   # After reading ray86 ------------------------------
   if $wk(read_status) {
      return
   } else {
      bar_message "$gbl(n_interface) interfaces were read."
   }
   after_read_model
}

# --------------------------------------------------------------------
# exec_read_ray86 ----------------------------------------------------
#            Read ray86 data file. (Extended Ray86 Version 3.0)
# --------------------------------------------------------------------
# Extended Ray86 Version 3.0 : P
#                        3.1 : P + SD
#                        3.2 : P + Poisson
#                        3.3 : P + Poisson + SD
# --------------------------------------------------------------------
# [arguments]
# data_file : Ray86 file name.
# --------------------------------------------------------------------
proc exec_read_ray86 {data_file} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   set fp [open $data_file]

   # version info 
   set line [get_line $fp]
   set version [lindex $line 0]

   # ------------------------------------------------------------
   # Read P-velocity model --------------------------------------

   # Read interface and node numers --------------------
   set line [get_line $fp]
   set gbl(n_interface) [lindex $line 0]
   set offset 1
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      if { [expr $i % 15] == 0 } {
	 set line [get_line $fp]
	 set offset 0
      }
      if $wk(read_status) return
      set gbl(n_node,$i) [lindex $line $offset]
      incr offset
   }
   # Read interface depth --------------------
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      foreach mode {x z} {
	 set gbl($mode,$i) [read_block $fp $gbl(n_node,$i)]
	 if $wk(read_status) return
      }
   }
   # Read speed network --------------------
   # Read max-velocity gradient, the others are neglected.
   set gbl(gradient_max) [lindex [get_line $fp] 3]

   # Model Bounds --------------------
   set data [split_line [get_line $fp]]
   foreach {mode i} {xmin 0 xmax 1 zmin 2 zmax 3} {
      set gbl($mode) [lindex $data $i]
   }
   # Velocity values --------------------
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      # Read node number
      set gbl(n_vnode,$i) [get_line $fp]
      # Read velocity grid
      foreach mode {vx vtop vbot} {
	 set gbl($mode,$i) [read_block $fp $gbl(n_vnode,$i)]
	 if $wk(read_status) return
      }
   }

   # ------------------------------------------------------------
   # Minor version ----------------------------------------------
   set minor_version [expr int($version*10) % 10]

   # ------------------------------------------------------------
   # Read Poisson's ratio ---------------------------------------
   # Initialize poisson's relation
   for {set i 1} {$i < $gbl(n_interface) } {incr i } {
      set gbl(poi,$i,top) ""
      set gbl(poi,$i,bot) ""
   }
   if { $minor_version >= 2 } {
      # read Poisson's mode
      set gbl(ERI,poisson) [lindex [get_line $fp] 0]
      if $wk(read_status) return
      switch $gbl(ERI,poisson) {
	 0 {
	    # None
	 }
	 1 {
	    # 1 value per 1 layer
	    set values [read_block $fp [expr $gbl(n_interface) - 1]]
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       set gbl(poi,$i,top) [lindex $values [expr $i - 1]]
	    }
	 }
	 2 {
	    # 1 value per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       set gbl(poi,$i,top) [read_block $fp $gbl(n_vnode,$i)]
	       if $wk(read_status) return
	    }
	 }
	 3 {
	    # 2 values per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       foreach mode {top bot} {
		  set gbl(poi,$i,$mode) [read_block $fp $gbl(n_vnode,$i)]
		  if $wk(read_status) return
	       }
	    }
	 }
	 default {
	    bar_message "Invalid Poisson's ratio mode.  Reading aborted."
	    return
	 }
      }
   }

   # ------------------------------------------------------------
   # Read SD of P-velocity --------------------------------------

   # Initialize
   for {set i 1} {$i < $gbl(n_interface) } {incr i } {
      set gbl(vtop,SD,$i) ""
      set gbl(vbot,SD,$i) ""
   }
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      set gbl(z,SD,$i) ""
   }
   if { $minor_version == 1 || $minor_version == 3 } {
      set gbl(ERI,inv) 1
      # Read Standard Deviation of velocity values
      for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	 foreach j {vtop vbot} {
	    set gbl($j,SD,$i) [read_block $fp $gbl(n_vnode,$i)]
	    if $wk(read_status) return
	 }
      }
      # Read  Standard Deviation of interface depth
      for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
	 set gbl(z,SD,$i) [read_block $fp $gbl(n_node,$i)]
	 if $wk(read_status) return
      }
   }

   # ------------------------------------------------------------
   # Read SD ratio, (SD of Vs)/(SD of Vp) -----------------------
   
   # Initialize
   for {set i 1} {$i < $gbl(n_interface) } {incr i } {
      set gbl(poi,SD,$i,top) ""
      set gbl(poi,SD,$i,bot) ""
   }
   if { $minor_version == 3 } {
      switch $gbl(ERI,poisson) {
	 0 {
	    # None
	 }
	 1 {
	    # 1 value per 1 layer
	    set values [read_block $fp [expr $gbl(n_interface) - 1]]
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       set gbl(poi,SD,$i,top) [lindex $values [expr $i - 1]]
	       if $wk(read_status) return
	    }
	 }
	 2 {
	    # 1 value per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       set gbl(poi,SD,$i,top) [read_block $fp $gbl(n_vnode,$i)]
	       if $wk(read_status) return
	    }
	 }
	 3 {
	    # 2 values per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       foreach mode {top bot} {
		  set gbl(poi,SD,$i,$mode) [read_block $fp $gbl(n_vnode,$i)]
		  if $wk(read_status) return
	       }
	    }
	 }
      }
   }
   close $fp

   # Set poisson's ratio relation
   set gbl(ERI,poisson_orig) $gbl(ERI,poisson)
}

# --------------------------------------------------------------------
# exec_read_old_ray86 ------------------------------------------------
#            Read ray86 data file. (Extended Ray86 Version 2.0 or former)
# --------------------------------------------------------------------
# [arguments]
# data_file : Ray86 file name.
# --------------------------------------------------------------------
proc exec_read_old_ray86 {data_file} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set fp [open $data_file]
   
   # Read interface and node numers --------------------
   set line [get_line $fp]
   set gbl(n_interface) [lindex $line 0]
   set offset 1
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      if { [expr $i % 15] == 0 } {
	 set line [get_line $fp]
	 set offset 0
      }
      set gbl(n_node,$i) [lindex $line $offset]
      incr offset
   }
   # Read interface depth --------------------
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      foreach mode {x z} {
	 vec_$mode$i clear
	 set gbl($mode,$i) [read_block $fp $gbl(n_node,$i)]
      }
   }

   # Read speed network --------------------
   # Read max-velocity gradient, the others are neglected.
   set gbl(gradient_max) [lindex [get_line $fp] 3]
   # Model Bounds --------------------
   set data [split_line [get_line $fp]]
   foreach {mode i} {xmin 0 xmax 1 zmin 2 zmax 3} {
      set gbl($mode) [lindex $data $i]
   }
   # Velocity strucutre --------------------
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      # Read node number
      set gbl(n_vnode,$i) [get_line $fp]
      # Read velocity grid
      foreach mode {vx vtop vbot} {
	 set gbl($mode,$i) [read_block $fp $gbl(n_vnode,$i)]
      }
   }
   #
   # For Extended-Ray86 Version 1 or later
   #
   # Read Standard Deviation of velocity values
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      foreach j {vtop vbot} {
	 set gbl($j,SD,$i) ""
      }
   }
   
   set flag 0
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      foreach j {vtop vbot} {
	 set data  [get_line $fp]
	 if { [eof $fp] } {
	    set flag 1
	    break;
	 }
	 set gbl($j,SD,$i) $data
      }
   }
   #
   # For Extended-Ray86 Version 2
   #
   # Read  Standard Deviation of interface depth
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      set gbl(z,SD,$i) ""
   }
   if { $flag != 1 } {
      for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
	 set data  [get_line $fp]
	 if { [eof $fp] } {
	    set flag 1
	    break;
	 }
	 set gbl(z,SD,$i) $data
      }
   }
   close $fp
   
   # poisson's ration relation
   set gbl(ERI,poisson) 0
   set gbl(ERI,poisson_orig) $gbl(ERI,poisson)
   initialize_poisson
}

# --------------------------------------------------------------------
# save_model ---------------------------------------------------------
#           Save model file in ray86 or v.in format
# --------------------------------------------------------------------
# [arguments]
# fmt : "ray86" or "v.in"
# --------------------------------------------------------------------
proc save_model {fmt} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   # Select output filename ------------------------------
   if { $fmt == "ray86"} {
      set types {
	 {"Ray86 files"	{.ray86}	}
	 {"All files"		*}
      }
      set ext ".ray86"
   } else {
      set types {
	 {"in files"	{.in}	}
	 {"dat files"	{.dat}	}
	 {"All files"		*}
      }
      set ext ".v_in"
   }
   set fn [tk_getSaveFile -filetypes $types -parent .  -defaultextension $ext]
   if { $fn == "" } {
      bar_message "Save canceled..."
      return -1
   }

   # Write ------------------------------
   if { $fmt == "ray86"} {
      write_ray86 $fn
   } else {
      write_vin $fn
   }
   bar_message "Wrote $gbl(n_interface) interfaces to $fn"
   update_titlebar
   backup_menu_update
}
# --------------------------------------------------------------------
# write_ray86 --------------------------------------------------------
#           Write ray86 data file (Extended-Ray86 V3 format)
# --------------------------------------------------------------------
# [arguments]
# fn : Filename to save.
# --------------------------------------------------------------------
proc write_ray86 {fn} {
   # Global variables --------------------------------------------------
   global gbl

   # ------------------------------------------------------------
   # Output Extended ray86 file version -------------------------
   set fp [open $fn w]
   set minor_version 0
   if { $gbl(ERI,inv) == 1 } {
      incr minor_version
   }
   if { $gbl(ERI,poisson) >= 1 } {
      set minor_version [ expr $minor_version + 2]
   }
   puts $fp "* Extended Ray86 file format version ----------"
   puts $fp [format "  3.%d" $minor_version]

   # ------------------------------------------------------------
   # Output numbers of interfaces and nodes ---------------------
   set tmp $gbl(n_interface)
   puts $fp "* Interface number and their nodes ----------"
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      set tmp "$tmp $gbl(n_node,$i)"
   }
   write_formated $fp $tmp 15 "5d"

   # ------------------------------------------------------------
   # Output interface depths ------------------------------------
   puts $fp "* Interfaces ------------------------------"
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      puts $fp "* Interface $i"
      write_formated $fp $gbl(x,$i) 10 "8.3f"
      write_formated $fp $gbl(z,$i) 10 "8.3f"
   }

   # Output Speed network  --------------------
   puts $fp "* ------------------------------\n* Speed Network"
   puts $fp [format "%5d%5d%10.3f%10.3f" -2 2 1.732 $gbl(gradient_max)]
   # Model Bounds --------------------
   puts $fp "* Model Bounds"
   puts $fp [format "%8.3f%8.3f%8.3f%8.3f"\
		 $gbl(xmin) $gbl(xmax) $gbl(zmin) $gbl(zmax)]
   
   # ------------------------------------------------------------
   # Output velocity grids  -------------------------------------
   puts $fp "* layers ------------------------------"
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      puts $fp "* Layer $i"
      write_formated $fp $gbl(n_vnode,$i) 1 "5d"
      write_formated $fp $gbl(vx,$i) 10 "8.3f"
      puts $fp "*       velocity of top side in layer $i"
      write_formated $fp $gbl(vtop,$i) 10 "8.3f"
      puts $fp "*       velocity of bottom side in layer $i"
      write_formated $fp $gbl(vbot,$i) 10 "8.3f"
   }


   # ------------------------------------------------------------
   # Output poisson's ratio -------------------------------------
   if { $gbl(ERI,poisson) >= 1} {
      puts $fp "* Poisson's ratio ------------------------------"
      puts $fp "* mode"
      puts $fp [format "%5d" $gbl(ERI,poisson)]
      if { $gbl(ERI,poisson) == 1 } {
	 # 1 value per 1 layer
	 set tmp {}
	 for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	    lappend tmp $gbl(poi,$i,top)
	 }
	 write_formated $fp $tmp 10 "8.3f"
      } else {
	 for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	    switch $gbl(ERI,poisson) {
	       2 {
		  # 1 value per 1 grid
		  puts $fp "* Layer $i"
		  for {set i 1} {$i < $gbl(n_interface)} {incr i} {
		     write_formated $fp $gbl(poi,$i,top) 10 "8.3f"
		  }
	       }
	       3 {
		  # 2 values per 1 grid
		  puts $fp "* Layer $i"
		  for {set i 1} {$i < $gbl(n_interface)} {incr i} {
		     foreach mode {top bot} {
			write_formated $fp $gbl(poi,$i,$mode) 10 "8.3f"
		     }
		  }
	       }
	    }
	 }
      }
   }

   # ------------------------------------------------------------
   # Output SD of P-velocity ------------------------------------
   if { $gbl(ERI,inv) == 1 } {
      puts $fp "* SD of P-Velocity ------------------------------"
      for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	 puts $fp "* Layer $i, top"
	 write_formated $fp $gbl(vtop,SD,$i) 10 "8.3f"
	 puts $fp "* Layer $i, bottom"
	 write_formated $fp $gbl(vbot,SD,$i) 10 "8.3f"
      }
      # Output interface depth standard deviations 
      puts $fp "* SD of Interface Depth ------------------------------"
      for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
	 puts $fp "* Interface $i"
	 write_formated $fp $gbl(z,SD,$i) 10 "8.3f"
      }
   }
   
   # ------------------------------------------------------------
   # Output SD ratio, (SD of Vs)/(SD of Vp) ---------------------
   if { $gbl(ERI,inv) == 1 && $gbl(ERI,poisson) >= 1 } {
      puts $fp "* SD ratio,  (SD of Vs)/(SD of Vp) -------------------"
      switch $gbl(ERI,poisson) {
	 0 {
	    # None
	    
	 }
	 1 {
	    # 1 value per 1 layer
	    set tmp {}
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       lappend tmp $gbl(poi,SD,$i,top)
	    }
	    write_formated $fp $tmp 10 "8.3f"
	 }
	 2 {
	    # 1 value per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       puts $fp "* layer $i"
	       write_formated $fp $gbl(poi,SD,$i,top) 10 "8.3f"
	    }
	 }
	 3 {
	    # 2 values per 1 grid
	    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	       foreach mode {top bot} {
		  puts $fp "* layer $i, $mode"
		  write_formated $fp $gbl(poi,SD,$i,$mode) 10 "8.3f"
	       }
	    }
	 }
      }
   }

   close $fp
}
# --------------------------------------------------------------------
# write_formated -----------------------------------------------------
#           Emulate "format" syntax of Fortran
#           num=10, mode=8.3f, ==> 10f8.3
# --------------------------------------------------------------------
# [arguments]
# fp    :  File pointer for output
# data  :  Data list to output
# num   :  Number of data in one row.
# mode  :  Output Format as "8.3f"
# --------------------------------------------------------------------
proc write_formated {fp data num mode} {
    set total [llength $data]
    set count 0
    while {1} {
	for {set i 0} {$i < $num} {incr i} {
	    puts -nonewline $fp [format "%$mode" [lindex $data $count]]
	    incr count
	    if {$count == $total} {
		break;
	    }
	}
	puts $fp ""
	if {$count == $total} {
	    break;
	}
    }
}

# --------------------------------------------------------------------
# get_line -----------------------------------------------------------
#            Read one line from a line.  Neglect comments.
# --------------------------------------------------------------------
# [arguments]
# fp : File pointer to output.
# --------------------------------------------------------------------
proc get_line {fp} {
    # Global variables --------------------------------------------------
    global wk

   # Main --------------------------------------------------
   set line "*"
   while {[string match {\**} $line]} {
      if { [eof $fp] } {
	 bar_message "Too short file.  Reading aborted."
	 set wk(read_status) 1
	 return
      }
      gets $fp line
   }
   return $line
}
# --------------------------------------------------------------------
# split_line ---------------------------------------------------------
#            Insert a blank in every eight characters and return it.
# --------------------------------------------------------------------
# [arguments]
# line : input strings.
# --------------------------------------------------------------------
proc split_line {line} {
    # Main --------------------------------------------------
    set split_line ""
    set len [string length "$line"]
    for {set i 0} {$i < $len} {set i [expr $i + 8]} {
	set split_line "$split_line [string range $line $i [expr $i+7]]"
    }
    return $split_line
}
# --------------------------------------------------------------------
# save_postscript ----------------------------------------------------
#            Save velocity structure in PostScript format.
# --------------------------------------------------------------------
proc save_postscript {} {
    # Global variables --------------------------------------------------
    global sys wk
    
    # Main --------------------------------------------------
    # Select Output filename.
    set types {
	{"ps files"	{.ps .eps}}
	{"All files"		*}
    }
    set ps_file [tk_getSaveFile -filetypes $types -parent .\
		-defaultextension .ps]
    if ![string compare $ps_file ""] {
	bar_message "Save PostScript canceled..."
	return
    }
    # Determine Save mode 
    set mode_list "color gray mono"
    set ans [tk_dialog .postscript "PostScript Mode" "Select PostScript Mode"\
		 question 1 color gray mono CANCEL]
    if {$ans == 3} {
	bar_message "Save PostScript canceled..."
	return
    }

    # Un-highlight all the interfaces
    set bak $wk(mark,size)
    set wk(mark,size) 1.5
    draw_model
    unselect_interface
    update

    # Output PostScript
    $sys(graph) postscript output $ps_file -landscape 1 -maxpect 1\
	    -decorations 0 -center 1 -colormode [lindex $mode_list $ans]\
	    -paperheight 29.7c -paperwidth 21c

    bar_message "Saved velocity structure in PostScript format."

    # recover
    set wk(mark,size) $bak
    draw_model
    selection_flash
}

# --------------------------------------------------------------------
# read_vin -----------------------------------------------------------
#            Read velocity structure in Zelt(v.in) format.
# --------------------------------------------------------------------
# [arguments]
# fn : v.in filename
# --------------------------------------------------------------------
proc read_vin {fn} {
    # Global variables --------------------------------------------------
    global gbl
    
    # Main --------------------------------------------------
    # Clear old structure --------------------
    clear_all

    # Open v.in file --------------------
    bar_message "Now reading the model file '$fn' ..."
    set fp [open $fn]

    # Read v.in --------------------
    set i 0
    while {1} {
	incr i
	# Initialize --------------------
	foreach mode {x z vx vtop vbot z,SD vtop,SD vbot,SD} {
	    set gbl($mode,$i) ""
	}
	# Read interface depths --------------------
	while {1} {
	    foreach mode {x z} {
		set data [split_vin [get_line $fp]]
		set n [llength $data]
		for {set k 1} {$k < $n} {incr k} {
		   lappend gbl($mode,$i) [lindex $data $k]
		}
	    }
	    if { [fblocked $fp] } {
		break;
	    }
	    get_line $fp
	    if {[lindex $data 0] != 1} {
		break
	    }
	}
	set gbl(n_node,$i) [llength $gbl(x,$i)]
	# End of file? (v.in)
	if { [eof $fp] } {
	    break;
	}
	# vx & vtop --------------------
	while {1} {
	    foreach mode {vx vtop} {
		set data [split_vin [get_line $fp]]
		set n [llength $data]
		for {set k 1} {$k < $n} {incr k} {
		    lappend gbl($mode,$i) [lindex $data $k]
		}
	    }
	    get_line $fp
	    if {[lindex $data 0] != 1} {
		break
	    }
	}
	set gbl(n_vnode,$i) [llength $gbl(vx,$i)]
	# vbot -----
	while {1} {
	    get_line $fp ; # Neglect vx (because ray86 does not has vx)
	    set data [split_vin [get_line $fp]]
	    set n [llength $data]
	    for {set k 1} {$k < $n} {incr k} {
		lappend gbl(vbot,$i) [lindex $data $k]
	    }
	    get_line $fp
	    if {[lindex $data 0] != 1} {
		break;
	    }
	}
    }
    # Finish --------------------
    set gbl(n_interface) $i
    bar_message "$gbl(n_interface) interfaces were read."
    close $fp

    # additional information
    initialize_poisson

    # Model Bounds
    set gbl(xmin) [lindex $gbl(x,1) 0]
    set gbl(xmax) [lindex $gbl(x,1) end]
    set gbl(zmin) [lindex [lsort -real $gbl(z,1)] 0]
    set gbl(zmax) [lindex [lsort -real $gbl(z,$gbl(n_interface))] end]
    # Set the max-velocity gradient (none-sense)
    set gbl(gradient_max) 3.0

    # after reading model
    after_read_model
}
# --------------------------------------------------------------------
# split_vin ----------------------------------------------------------
#            Insert a blank in every 8 characters for v.in
# --------------------------------------------------------------------
# [arguments]
# line : input strings.
# --------------------------------------------------------------------
proc split_vin {line} {
    # Main --------------------------------------------------
    set len [string length "$line"]
    set split_line [string range $line 0 2]
    for {set i 3} {$i < $len} {set i [expr $i + 7]} {
	set split_line "$split_line [string range $line $i [expr $i+6]]"
    }
    return $split_line
}
# --------------------------------------------------------------------
# check_model_vin ----------------------------------------------------
#           Ray86 and v.in formats are not completely equivalent in math.
#           v.in is superior to Ray86 to describe velocity models.
#           This procedure translates v.in into Ray86 by force.
# --------------------------------------------------------------------
proc check_model_vin {} {
    # Global variables --------------------------------------------------
    global gbl

    # Main --------------------------------------------------
    # Complement of left end of the structure  --------------------
    for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
	# Complement in case of left-end != xmin (interface point)
	if { [lindex $gbl(x,$i) 0] != $gbl(xmin) } {
	    set gbl(x,$i) [linsert $gbl(x,$i) 0 $gbl(xmin)]
	    set gbl(z,$i) [linsert $gbl(z,$i) 0 [lindex $gbl(z,$i) 0]]
	    set gbl(n_node,$i) [llength $gbl(x,$i)]
	}
	# Complement in case of left-end != xmin (velocity grid)
	if {$i < $gbl(n_interface) && [lindex $gbl(vx,$i) 0] != $gbl(xmin)} {
	    set gbl(vx,$i) [linsert $gbl(vx,$i) 0 $gbl(xmin)]
	    set gbl(vtop,$i) [linsert $gbl(vtop,$i) 0 [lindex $gbl(vtop,$i) 0]]
	    set gbl(vbot,$i) [linsert $gbl(vbot,$i) 0 [lindex $gbl(vbot,$i) 0]]
	    set gbl(n_node,$i) [llength $gbl(vx,$i)]
	}
    }
    # Complement of velocity grid --------------------
    for {set i 1} {$i < $gbl(n_interface)} {incr i} {
	if { [llength $gbl(vtop,$i)] != [llength $gbl(vbot,$i)] } {
	    # Generate vbot
	    puts stdout "DEBUG: force velocity, $i layer, bottom $gbl(vbot,$i)"
	    set vtmp [lindex $gbl(vbot,$i) 0]
	    set gbl(vbot,$i) ""
	    for {set j 0} {$j < $gbl(n_vnode,$i)} {incr j} {
		lappend gbl(vbot,$i) $vtmp
	    }
	}
    }
}
# --------------------------------------------------------------------
# model_from_netcdf --------------------------------------------------
#           Create new model from a netCDF file
# --------------------------------------------------------------------
proc model_from_netcdf {} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set types {
      {"netCDF files"	{.grd}	}
      {"All files"	*}
   }
   set netcdf_file [tk_getOpenFile -filetypes $types -parent .]
   if ![string compare $netcdf_file ""] {
      # When canceled.
      bar_message "No file is selected."
      return
   }
   # read netCDF info
   bar_message "Creating new model from \"$netcdf_file\" ..."
   set f [open "| grdinfo -C $netcdf_file"]
   scan [gets $f] "%*s%f%f%f%f" gbl(xmin) gbl(xmax) gbl(zmin) gbl(zmax)
   close $f
   # create a rectangle model
   new_model
   # add an interface (flat)
   insert_interface down
   # setup to use the netCDF as the velocity field
   set gbl(hamen,mformat) 3
   set gbl(file,cdffile) $netcdf_file
   set gbl(option,backend) "hamen"
   set gbl(option,label) 0
   draw_grid
   bar_message "New model is created from \"$netcdf_file\""
   # invoke parameter window and show message dialog
   HAMEN::parameter_window
   .hamen_parameter.tabset select end
   tk_messageBox -icon info -type ok\
       -message "Velocity field is read from the netCDF\n\nYou can specify background image."
}
# --------------------------------------------------------------------
# initialize_SD ------------------------------------------------------
#            Initialize Standard Deviation values
# --------------------------------------------------------------------
proc initialize_SD {} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   # Velocity value
   for {set i 1} {$i < $gbl(n_interface)} {incr i} {
      foreach v {vtop vbot} {
	 set n [llength $gbl($v,SD,$i)]
	 if {$n >= $gbl(n_vnode,$i)} {
	    set gbl($v,SD,$i) [lrange $gbl($v,SD,$i) 0\
				   [expr $gbl(n_vnode,$i) -1]]
	 } else {
	    for {set j $n} {$j < $gbl(n_vnode,$i)} {incr j} {
	       lappend gbl($v,SD,$i) 0.0
	    }
	 }
      }
   }
   # Interface depth
   for {set i 1} {$i <= $gbl(n_interface)} {incr i} {
      set n [llength $gbl(z,SD,$i)]
      if {$n >= $gbl(n_node,$i)} {
	 set gbl(z,SD,$i) [lrange $gbl(z,SD,$i) 0 [expr $gbl(n_node,$i) -1]]
      } else {
	 for {set j $n} {$j < $gbl(n_node,$i)} {incr j} {
	    lappend gbl(z,SD,$i) 0.0
	 }
      }
   }
}
# --------------------------------------------------------------------
# check_ray86_version ------------------------------------------------
#           Check ray86 file version
# --------------------------------------------------------------------
# [arguments]
# data_file : Ray86 file name.
# --------------------------------------------------------------------
proc check_ray86_version {data_file} {
   set fp [open $data_file]
   set line [get_line $fp]
   close $fp
   if { [llength $line] > 1 } {
      # older than version 2
      return 2
   }
   if { [llength $line] == 1 } {
      # version 3 or later
      return [lindex $line 0]
   }
   return -1
}

# --------------------------------------------------------------------
# read_block ---------------------------------------------------------
#           Read one block
# --------------------------------------------------------------------
# [arguments]
# fp : file pointer
# n  : number of data to read.
# --------------------------------------------------------------------
proc read_block {fp n} {
   # Global variables --------------------------------------------------
   global wk
   
   # Main --------------------------------------------------
   set result {}
   set i 0
   while {$i < $n} {
      if [eof $fp] { 
	 set wk(read_status) 1
	 return 
      }
      set data [split_line [get_line $fp]]
      set result [concat $result $data]
      set i [expr $i + [llength $data]]
   }
   return $result
}

# --------------------------------------------------------------------
# after_read_model ---------------------------------------------------
#           Processes after reading model
# --------------------------------------------------------------------
proc after_read_model {} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # standard deviation adjustment
   initialize_SD
   
   # poisson's ratio adjustment
   adjust_poisson

   # Check for v.in (in cace of converting from v.in into ray86)
   check_model_vin
    
   # Select an interface
   set wk(selected) [expr int($gbl(n_interface)/2)]

   # Draw model ------------------------------
   bar_message "Now drawing the model..."
   draw_frame
   draw_model
   update_bindings
   bar_message "Here we go!"
}