# --------------------------------------------------------------------
# discriminate_pickfile ----------------------------------------------
#            Discriminate pickfile format, wheter "macray" or "zelt"
# --------------------------------------------------------------------
# [arguments]
# file :  pick file in either "macray" and "zelt" format
# --------------------------------------------------------------------
proc discriminate_pickfile {file} {
    # readable check
    if {[string length $file] < 1} { return none }
    if {![file readable $file]} {
	tk_messageBox -message "Pick file \"$file\" is not readable."\
	    -type ok -icon error
	return none
    }
    # open and read first line
    set fp [open $file]
    gets $fp line
    close $fp
    # check type
    if {[lindex $line 3] == 0} {
	# zelt format   (4-th component must be zero)
	return zelt
    } else {
	# macray format (4-th component is distance multiplier, can't be zero)
	return macray
    }
    return none
}
# --------------------------------------------------------------------
# tx_in_append_zelt --------------------------------------------------
#            Append traveltime pick data to tx.in from zelt format.
# --------------------------------------------------------------------
# [arguments]
# out  :  File pointer to output.
# file :  pick file in "zelt" format
# --------------------------------------------------------------------
proc tx_in_append_zelt {out file} {
   # Only neglect the end of the data line, "0 0 0 -1"
   set fp [open $file "r"]
   while { [gets $fp line] != -1} {
      if { [llength $line] != 4 } { continue }
      if { [lindex $line 0] != 0.0 ||\
	       [lindex $line 1] != 0.0 ||\
	       [lindex $line 2] != 0.0 ||\
	       [lindex $line 3] != -1 } {
	 puts $out $line
      }
   }
   close $fp
}
# --------------------------------------------------------------------
# tx_in_append_macray ------------------------------------------------
#            Append traveltime pick data to tx.in from macray format.
# --------------------------------------------------------------------
# [arguments]
# out  :  File pointer to output.
# file :  pick file in "macray" format
# acc  :  default accuracy
# --------------------------------------------------------------------
proc tx_in_append_macray {out file acc} {
   # Convert macray into zelt ------------------------------
   set fp [open $file "r"]
   
   # read first line
   gets $fp line
   set shot_range [lindex $line 0]
   set depth [lindex $line 1]
   set time_offset [lindex $line 2]
   # warning: now, dist_multi is not used.
   set dist_multi [lindex $line 3]

   # Read data and convert and output them
   set n 0
   while {[gets $fp line] != -1} {
      if { [llength $line] != 5 || [regexp "\#" $line] } {
	 continue
      }
      set dist($n) [lindex $line 1]
      set time($n) [lindex $line 2]
      set accuracy($n) [lindex $line 3]
      if { $accuracy($n) <= 0.0 } {
	 set accuracy($n) $acc
      }
      set type($n) [lindex $line 4]
      incr n
   }
   close $fp

   # write out right hand side
   puts $out [format "%10.3f%10.3f%10.3f%10d" $shot_range 1 0.0 0]
   for {set i 0} {$i < $n} {incr i} {
      if { $dist($i) >= 0.0 } {
	 puts $out [format "%10.3f%10.3f%10.3f%10d"\
			[expr $dist($i) + $shot_range]\
			$time($i) $accuracy($i) $type($i)]
      }
   }

   # write out left hand side
   puts $out [format "%10.3f%10.3f%10.3f%10d" $shot_range -1 0.0 0]
   for {set i 0} {$i < $n} {incr i} {
      if { $dist($i) < 0.0 } {
	 puts $out [format "%10.3f%10.3f%10.3f%10d"\
			[expr $dist($i) + $shot_range]\
			$time($i) $accuracy($i) $type($i)]
      }
   }
}
# --------------------------------------------------------------------
# tx_in_append_auto_2pt ----------------------------------------------
#            Append traveltime pick data to tx.in
#            Data are automatically generated for 2pt-raytracing.
# --------------------------------------------------------------------
# [arguments]
# out  :  File pointer to output.
# rx   :  receiver x-coordinate (km), i.e. shot range.
# --------------------------------------------------------------------
proc tx_in_append_auto_2pt {out rx} {
   global gbl
   if { $gbl(trace,dx2pt) <= 0.0 } return
   # right side
   puts $out [format "%10.3f%10.3f%10.3f%10d" $rx 1.0 0.0 0]
   foreach i [split $gbl(zelt,invr,ivray) ","] {
      for { set x $gbl(xmin) } { $x < $gbl(xmax) } \
	  { set x [expr $x + $gbl(trace,dx2pt)] } {
	     if { $rx <= $x } {
		puts $out [format "%10.3f%10.3f%10.3f%10d" $x 0.0 0.0 $i]
	     }
	  }
   }
   # left side
   puts $out [format "%10.3f%10.3f%10.3f%10d" $rx -1.0 0.0 0]
   foreach i [split $gbl(zelt,invr,ivray) ","] {
      for { set x [expr $rx - $gbl(trace,dx2pt)] } { $x > $gbl(xmin) } \
	  { set x [expr $x - $gbl(trace,dx2pt)] } {
	     if { $x > $gbl(xmin) } {
		puts $out [format "%10.3f%10.3f%10.3f%10d" $x 0.0 0.0 $i]
	     }
	  }
   }
}
