# --------------------------------------------------------------------
# TX namespace 
#  
# -  TX::plot $fn
#         plot tx graph.  $fn is a synthetic traveltime data file
#         $fn is either in "macray(pasteup)" or "tx.out" format
#
# -  TX::show_position %x %y
#         show cursor position in the tx graph window
#
# -  TX::highlight
#         highlight one raypath corresponding to the selected synthetic
#         traveltime symbol.
#
# -  TX::save_postsript
#         save tx graph in a postscript file
#
# -  TX::select_axis
#         select and edit an axis of tx-graph
#
# --------------------------------------------------------------------
namespace eval TX {
   namespace export *

   # --------------------------------------------------------------------
   # Plot T-X graph inside modeling window both observed and synthetics
   # --------------------------------------------------------------------
   #     fn : synthetic traveltime data file name
   # --------------------------------------------------------------------
   proc plot {fn} {
      # Global variables
      global gbl sys
      
      # Main --------------------------------------------------
      bar_message "Now drawing T-X graph ...."

      # Clear all the old graph
      eval {$sys(tx) element delete} [$sys(tx) element names]
      eval {$sys(tx) marker delete} [$sys(tx) marker names]

      # set axis direction
      $sys(tx) yaxis conf -descending $gbl(plot,taxis_dir)
      $sys(tx) y2axis conf -descending $gbl(plot,taxis_dir)
      
      # Synthetics
      if [file readable $fn] { plot_synthetics $fn }

      # Observed
      for {set i 1} {$i <= $gbl(src,n_source)} {incr i} {
	 if {$gbl(src,$i,onoff)} {
	    plot_observed $gbl(src,$i,pick)
	 }
      }

      # Accessory
      if { $gbl(plot,vred) > 0.0 } {
	 $sys(tx) yaxis configure\
	     -title [format "T - X/%g (sec)" $gbl(plot,vred)]
      } else {
	 $sys(tx) yaxis configure -title "T(sec)"
      }
      set trange [$sys(tx) yaxis limit]
      $sys(tx) y2axis conf -min [lindex $trange 0] -max [lindex $trange 1]
   }

   # --------------------------------------------------------------------
   # plot synthetics inside modeling
   # --------------------------------------------------------------------
   proc plot_synthetics {fn} {
      # Global variables
      global gbl sys

      # main ------------------------------
      if {$gbl(plot,tx) == 2} {
	 # plot symbol mode
	 $sys(tx) element create synthetic -pixels 2 -linewidth 0\
	     -fill black -color black
      }

      # descriminating data file format
      set fp [open $fn]
      if { [gets $fp line1] < 0 || [gets $fp line2] < 0 } { 
	 # too short file
	 close $fp
	 return 
      }
      close $fp
      set n1 [llength $line1]
      switch -- $n1 {
	 2 {
	    # (x,t)     : travetltime at surface
	    plot_syn_surface $fn
	 }
	 3 {
	    # (x, z, t) : traveltime field
	    plot_syn_tfield $fn
	 }
	 4 {
	    set n2 [llength $line2]
	    if { $n2 == 5 } {
	       # macray
	       plot_syn_macray $fn
	    } else {
	       # probably, txout
	       plot_syn_txout $fn
	    }
	 }
      }
   }

   # --------------------------------------------------------------------
   # plot synthetics inside modeling window using tx.in format data
   # --------------------------------------------------------------------
   proc plot_syn_txout {fn} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      # Synthetic part ------------------------------
      set prev_type -1
      set station_x 0.0
      set coords {}

      # plot ----------
      set fp [open $fn]
      while { [gets $fp line] != -1 } {
	 set dist [lindex $line 0]
	 set time [lindex $line 1]
	 set accuracy [lindex $line 2]
	 set type [lindex $line 3]
	 # The end of data?
	 if { $dist == 0.0 && $time == 0.0 && $accuracy == 0.0 && $type == 1 } {
	    synthetic_branch $station_x $coords
	    set coords {}
	    break
	 }
	 # Station change?
	 if {$type <= 0} {
	    synthetic_branch $station_x $coords
	    set station_x $dist
	    set coords {}
	    continue
	 }
	 # Wave type change?
	 if {$type != $prev_type} {
	    synthetic_branch $station_x $coords
	    set coords {}
	    set prev_type $type
	 }
	 lappend coords "$dist $time"
      }
      close $fp
   }

   # --------------------------------------------------------------------
   # plot synthetics using tfield data
   # --------------------------------------------------------------------
   proc plot_syn_tfield {fn} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      tk_messageBox -icon warning -type ok\
	  -message "Traveltime field data cannot be plot."
      return
   }

   # --------------------------------------------------------------------
   # plot synthetics using macray (pasteup) format data
   # --------------------------------------------------------------------
   proc plot_syn_macray {fn} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      set prevtype -30000
      set fp [open $fn "r"]
      gets $fp line
      set sx [lindex $line 0]
      set coords {}
      while {[gets $fp line] != -1} {
	 set cx [expr [lindex $line 1] + $sx]
	 set ct [lindex $line 2]
	 set type [lindex $line 4]
	 if { $prevtype != $type } {
	    set prevtype $type
	    if {[llength $coords] > 1} {
	       synthetic_branch $sx $coords
	    }
	    set coords {}
	 }
	 lappend coords "$cx $ct"
      }
      close $fp
      if {[llength $coords] > 1} {
	 synthetic_branch $sx $coords
      }
   }

   # --------------------------------------------------------------------
   # plot synthetics using (x, t) data
   # --------------------------------------------------------------------
   proc plot_syn_surface {} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      # determine sx
      set sx 0
      for {set i 1} {$i <= $gbl(src,n_source)} {incr i} {
	 if {$gbl(src,$i,onoff)} {
	    set sx $gbl(src,$i,x)
	 }
      }
      # read
      set fp [open $fn "r"]
      set prevx $gbl(xmax)
      set coords {}
      while {[gets $fp line] != -1} {
	 set cx [lindex $line 0]
	 set ct [lindex $line 1]
	 if { $cx < $prevx } {
	    set prevx $cx
	    if { [llength $coords] > 1 } {
	       # plot
	       synthetic_branch $sx $coords
	    }
	    set coords {}
	 }
	 set prevx $cx
	 lappend coords "$cx $ct"
      }
      close $fp
      if { [llength $coords] > 1 } {
	 # plot
	 synthetic_branch $sx $coords
      }
   }

   # --------------------------------------------------------------------
   # plot synthetic T-X graph (1-branch)
   # --------------------------------------------------------------------
   # [arguments]
   # x   :  station position in km
   # c   :  (x1,t1, x2,t2, ... xn,tn)  x is absolute distance.
   # --------------------------------------------------------------------
   proc synthetic_branch {x c} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      if {[llength $c] < 1} {
	 return
      }
      # Reduction
      if { $gbl(plot,vred) > 0.0 } {
	 set v [expr $gbl(plot,vred) * 1.0]
	 set coord {}
	 foreach {dist time} [join $c] {
	    lappend coord "$dist [expr $time-abs($dist-$x)/$gbl(plot,vred)]"
	 }
      } else {
	 set coord $c
      }
      # plot
      switch -- $gbl(plot,tx) {
	 1 {
	    # plot with line
	    if {[llength $c] < 2} return
	    set id [expr [lindex [lsort -integer [$sys(tx) element names]] end] +1]
	    $sys(tx) element create $id -data [join $coord]\
		-pixels 0 -linewidth 1 -color black
	 }
	 2 {
	    # plot with symbols
	    set previous [$sys(tx) element cget synthetic -data]
	    $sys(tx) element configure synthetic -data [join "$previous $coord"]
	 }
	 default {
	    # No synthetics
	 }
      }
   }

   # --------------------------------------------------------------------
   # plot observed traveltime data inside modeling
   # --------------------------------------------------------------------
   proc plot_observed {fn} {
      switch -- [discriminate_pickfile $fn] {
	 "zelt" {
	    plot_observed_txin $fn
	 }
	 "macray" {
	    plot_observed_macray $fn
	 }
	 "none" {
	    # not appropriate pick file
	 }
      }
   }

   # --------------------------------------------------------------------
   # plot observed traveltime data in tx.in format
   # --------------------------------------------------------------------
   # fn   :  Input filename
   # --------------------------------------------------------------------
   proc plot_observed_txin {fn} {
      # Global variables --------------------------------------------------
      global sys gbl
      
      # Main --------------------------------------------------
      # File check ------------------------------
      if ![file exists $fn] {
	 return
      }
      # Open and read it ------------------------------
      set fp [open $fn "r"]
      gets $fp line
      set station_x 0.0
      # Get station position
      set station_x [lindex $line 0]
      # Color procedure ----------
      switch -- $gbl(plot,pick_color) {
	 1 {
	    # Pick code
	    proc color {t n} {
	       global gbl sys
	       return [lindex $sys(colors) [expr ($t-1) % $sys(ncolor)]]
	    }
	 }
	 2 {
	    # Shot(OBS)
	    proc color {t n} {
	       global gbl sys
	       return [lindex $sys(colors) [expr $n % $sys(ncolor)]]
	    }

	 }
	 default {
	    # None color(black)
	    proc color {t n} {return "black"}
	 }
      }
      
      # Loop ----------
      set n -1;  # station number
      set count 1
      while { [gets $fp line] != -1} { 
	 if { [llength $line] != 4 } { continue }
	 if { $count != 1 } {
	    incr count -1
	    continue
	 }
	 set count $gbl(param,plot,observe_thinout)
	 set dist [lindex $line 0]
	 set time [lindex $line 1]
	 set acc  [lindex $line 2]
	 set type [lindex $line 3]
	 if { $type == 0 } {
	    set station_x $dist
	    incr n
	    continue
	 }
	 plot_1_traveltime $station_x $dist $time $acc [color $type $n]
      }
      close $fp
   }

   # --------------------------------------------------------------------
   # plot observed traveltime data in macray (pasteup) format 
   # --------------------------------------------------------------------
   # fn   :  Input filename
   # --------------------------------------------------------------------
   proc plot_observed_macray {fn} {
      # Global variables --------------------------------------------------
      global sys gbl
      
      # Main --------------------------------------------------
      # File check ------------------------------
      if ![file exists $fn] {
	 return
      }
      # Open and read it ------------------------------
      set fp [open $fn "r"]
      gets $fp line
      # Get station position
      set station_x [lindex $line 0]
      # Color procedure ----------
      switch -- $gbl(plot,pick_color) {
	 1 {
	    # Pick code
	    proc color {t} {
	       global gbl sys
	       return [lindex $sys(colors) [expr ($t-1) % $sys(ncolor)]]
	    }
	 }
	 2 {
	    # Shot(OBS)
	    set i [lsearch [make_source_list] [format "%.3f" $station_x]]
	    set c [lindex $sys(colors) [expr $i % $sys(ncolor)]]
	    eval "proc color {t} {return $c}"
	 }
	 default {
	    # None color(black)
	    proc color {t} {return "black"}
	 }
      }
      
      # Loop ----------
      set count 1
      while {[gets $fp line] != -1} {
	 if { [llength $line] != 5 || [regexp "\#" $line] || $count != 1} {
	    incr count -1
	    continue
	 }
	 set count $gbl(plot,observe_thinout)
	 set x [lindex $line 1]
	 set t [lindex $line 2]
	 set x [expr $x + $station_x]
	 set a [lindex $line 3]
	 set type [lindex $line 4]
	 if { $a <= 0.0 } {
	    set a $gbl(plot,pick_acc)
	 }
	 plot_1_traveltime $station_x $x $t $a [color $type]
      }
      close $fp
   }


   # --------------------------------------------------------------------
   # plot 1 traveltime mark (observed)
   # --------------------------------------------------------------------
   # [arguments]
   # x  : station position in km.
   # d  : pick position in km (absolute, not relative)
   # t  : traveltime in second.
   # a  : accuracy in second.
   # cl : color
   # --------------------------------------------------------------------
   proc plot_1_traveltime {x d t a cl} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      # Reduce
      if { $gbl(plot,vred) > 0.0 } {
	 set t [expr $t - abs($d - $x)/$gbl(plot,vred)]
      }
      # Plot
      set c [list $d [expr $t-$a] $d [expr $t+$a]]
      $sys(tx) marker create line -coords $c -outline $cl
   }

   # --------------------------------------------------------------------
   # Show the mouse corsor position in the TX window
   # --------------------------------------------------------------------
   #  x  :  Mouse postion
   #  y  :  Mouse position
   # --------------------------------------------------------------------
   proc show_position {x y} {
      # Global variables
      global sys wk

      # Main --------------------------------------------------
      set xz [$sys(tx) invtransform $x $y]
      set gx [lindex $xz 0]
      set gz [lindex $xz 1]
      catch {$wk(w,pos).message configure\
		 -text [format "(%8.3fkm, %8.3fsec)" $gx $gz]}
   }
   # --------------------------------------------------------------------
   # Highlight traveltime symbols and its corresponding raypath.
   # --------------------------------------------------------------------
   #  x  :  Mouse postion
   #  y  :  Mouse position
   # --------------------------------------------------------------------
   proc highlight {x y} {
      # Global variables
      global gbl sys
      
      # Main --------------------------------------------------
      # Delete previous closest mark
      foreach i [$sys(graph) element names] {
	 $sys(graph) element deactivate $i
      }
      catch {$sys(tx) element delete activemark}
      
      # Check plot mode ----------
      if { [$sys(tx) element names synthetic] == ""} {
	 bar_message "You must plot with symbols to highlight synthetics."
	 return
      }

      # Search nearest point and highlight it. ----------
      if {[$sys(tx) element closest $x $y closest -halo 5] == 0} {
	 bar_message "No near synthetic."
	 return
      }
      bar_message [format "Closest point is (%.3fkm,%3.fsec)"\
		       $closest(x) $closest(y)]
      $sys(tx) element create activemark -data "$closest(x) $closest(y)"
      $sys(tx) element activate activemark
      
      # Search corresponding raypath and highlight it ----------
      if { ![$sys(graph) element exist ray(0)] } {
	 return
      }
      set surface [$sys(graph) transform $closest(x) 0.0]
      set x [lindex $surface 0]
      set y [lindex $surface 1]
      if {[$sys(graph) element closest $x $y closest -halo 1] == 0 || \
	      ![string match "ray(*)" $closest(name)]  } {
	 bar_message "ERROR!?: No corresponding raypath."
	 return
      }

      $sys(graph) element activate $closest(name)
   }

   # --------------------------------------------------------------------
   # save T-X graph as a PostScript file
   # --------------------------------------------------------------------
   proc save_postscript {} {
      # Global variables --------------------------------------------------
      global gbl sys
      
      # Main --------------------------------------------------
      # Select Output filename.
      set types {
	 {"ps files"	{.ps .eps}}
	 {"All files"		*}
      }
      set ps_file [tk_getSaveFile -filetypes $types -parent .\
		       -defaultextension .ps]
      if ![string compare $ps_file ""] {
	 bar_message "Save PostScript T-X canceled..."
	 return
      }
      # Determine Save mode 
      set mode_list "color gray mono"
      set ans [tk_dialog .postscript "PostScript Mode" "Select PostScript Mode"\
		   question 1 color gray mono CANCEL]
      if {$ans == 3} {
	 bar_message "Save PostScript T-X canceled..."
	 return
      }

      # Output PostScript
      $sys(tx) postscript output $ps_file -landscape 1 -maxpect 1\
	  -decorations 0 -center 1 -colormode [lindex $mode_list $ans]\
	  -paperheight 29.7c -paperwidth 21c

      bar_message "Saved T-X graph in PostScript format."
   }

   # --------------------------------------------------------------------
   # select and edit t-axis
   # --------------------------------------------------------------------
   #   x   : x coordinate of mouse cursor. (Graph widget coordiantes.
   #   y   : y coordinate of mouse cursor. (Graph widget coordiantes.  
   #   XX  : x coordinate of mouse cursor. (root window coordiantes.)  
   #   YY  : y coordinate of mouse cursor. (root window coordiantes.)
   # --------------------------------------------------------------------
   proc select_axis {x y XX YY} {
      # Global variable ------------------------------------------------
      global gbl sys

      # Main -----------------------------------------------------------
      # Inside graph area?
      if { [$sys(tx) inside $x $y] } {
	 # Inside graph area.
	 return
      }
      # Range of axis
      set xx [$sys(tx) axis limits x]
      set x1 [lindex $xx 0]
      set x2 [lindex $xx 1]
      set tt [$sys(tx) axis limits y]
      set t1 [lindex $tt 0]
      set t2 [lindex $tt 1]

      # Get cursor postion.
      set xt [$sys(tx) invtransform $x $y]
      set x [lindex $xt 0]
      set t [lindex $xt 1]
      
      # Determine axis.
      if { $x < $x1 || $x > $x2} {
	 # t-axis area
	 set_axis t $XX $YY
      } else {
	 # x-axis area
	 set_axis x $XX $YY
      }
   }

}