# --------------------------------------------------------------------
# select_grid --------------------------------------------------------
#            Select velocity grid closest to mouse cursor
# --------------------------------------------------------------------
# [arguments] 
#   x    :  x coordinate of mouse cursor. (Graph widget coordiantes.
#   y    :  y coordinate of mouse cursor. (Graph widget coordiantes.
#   XX   :  x coordinate of mouse cursor. (root window coordiantes.)
#   YY   :  y coordinate of mouse cursor. (root window coordiantes.)
#   mode :  Process after selection. "delete", "edit" or "move"
# --------------------------------------------------------------------
proc select_grid {x y XX YY mode} {
    # Global variables --------------------------------------------------
    global gbl sys wk
    
    # Main --------------------------------------------------
    # Initialize
    set gbl(move_grid,original_name) ""

    # When invisible grid, return.
    if {$gbl(option,grid) == 0} return
    
    # Get mouse position ------------------------------
    set xz [$sys(graph) invtransform $x $y]
    set gx [lindex $xz 0]
    set gz [lindex $xz 1]

    # Select interface
    unselect_interface
    set wk(selected) [search_layer $gx $gz]
    selection_flash
    # Left and right end
    if {$gx < $gbl(xmin) || $gx > $gbl(xmax)} {
	return
    }
    # Upper and lower end
    if {$wk(selected) < $gbl(n_interface)} {
	set top [interpolate_depth $wk(selected) $gx]
	set bottom [interpolate_depth [expr $wk(selected) + 1] $gx]
    } else {
	return
    }
    # Mouse cursor is valid place?
    if {[expr ($bottom - $gz)*($gz - $top)] < 0} {
	return
    }
    # Compute closest grid --------------------
    # List up all the grid in the layer
    set marker_list [$sys(graph) marker names m($wk(selected),*,v)]
    set near_name ""
    set nearest 1.0e30
    foreach i $marker_list {
	# Get the grid x-coordinate
	set coords [$sys(graph) marker cget $i -coords]
	# Nearest grid in this layer
	if { [lindex $coords 1] <= $gz && [lindex $coords 3] >= $gz} {
	    set dist [expr abs($gx - [lindex $coords 0])]
	    if {$dist < $nearest} {
		set nearest $dist
		set near_name $i
	    }
	}
    }
    # Enough close?
    if { $near_name != "" } {
       set coords [$sys(graph) marker cget $near_name -coords]
       set window_x [lindex [$sys(graph) transform\
				 [lindex $coords 0] [lindex $coords 1]] 0]
       if { [expr abs($x - $window_x)] > $wk(halo)} {
	  bar_message "Can't select any velocity grid."
	  return
       }
       # Highlight the velocity grid
       $sys(graph) marker create line -name $near_name.tmp \
	   -coords [$sys(graph) marker cget $near_name -coords]\
	   -linewidth 2 -outline $wk(col,select)
       # Message
       bar_message "A velocity grid in the $wk(selected)-Layer is selected."
    }
    # Mode switch
    if ![string compare $mode "edit"] {
	# Edit mode
	edit_velocity $XX $YY $near_name
    } elseif { ![string compare $mode "delete"] } {
	# Delete mode
	delete_velocity $near_name
    } else {
	# Begin to move mode
	move_grid_start $near_name $gx
    }
}

# --------------------------------------------------------------------
# select_neighbor_vgrid ----------------------------------------------
#            select neighbor vgrid
# --------------------------------------------------------------------
#   d    : direction, -1:left, +1:right
#   layer: layer number
#   vnode: v-grid number from the left
# --------------------------------------------------------------------
proc select_neighbor_vgrid {d layer vnode} {
   # Global variables --------------------------------------------------
   global gbl sys wk
   
   # Main --------------------------------------------------
   set newvnode [expr $vnode + $d]
   if {$newvnode < 0} {
      # left end
      return
   }
   if {$newvnode > [expr [llength $gbl(vx,$layer)] -1] } {
      # right end
      return
   }
   
   # apply prev ----------
   finish_velocity $layer $vnode ok .velocity_window

   # un-highlight prev
   $sys(graph) marker delete m($layer,$vnode,v).tmp

   # select new vnode ----------
   # highlight new
   set newname m($layer,$newvnode,v)
   set coord [$sys(graph) marker cget $newname -coords]
   $sys(graph) marker create line -name $newname.tmp \
       -coords $coord -linewidth 2 -outline $wk(col,select)

   # get position in window coordinate system (mouse poistion)
   set gx [lindex $coord 0]
   set gy [expr ([lindex $coord 1] + [lindex $coord 3])/2.0]
   set xz [$sys(graph) transform $gx $gy]
   edit_velocity [lindex $xz 0] [lindex $xz 1] $newname
}

# --------------------------------------------------------------------
# edit_velocity ----------------------------------------------------
#            Velocity grid edit window
# --------------------------------------------------------------------
#   XX   : x coordinate of mouse cursor. (root window coordiantes.)
#   YY   : y coordinate of mouse cursor. (root window coordiantes.)
#   name : Marker name of velocity grid.
# --------------------------------------------------------------------
proc edit_velocity {XX YY name} {
   # Global variables --------------------------------------------------
   global sys gbl wk
   
   # Main --------------------------------------------------
   # Get the layer number and node number ----------
   set tmp [split [string trim $name "m(,v)"] ","]
   set layer [lindex $tmp 0]
   set vnode [lindex $tmp 1]

   # Invoke window ------------------------------
   set vw .velocity_window
   if [winfo exist $vw] {
      regsub \\d*x\\d* [wm geometry $vw] "" geometry
      destroy $vw
   } else {
      set geometry [format "+%d+%d" [expr $XX + 2] $YY]
   }
   toplevel $vw
   wm geometry $vw $geometry
   
   # Title frame ------------------------------
   title_frame $vw.title "Velocity Grid : Layer $wk(selected)"

   # Grid location frame ------------------------------
   set vwl $vw.location
   frame $vwl -relief ridge -bd 2
   label $vwl.xlabel -text "Grid position.   X = "
   entry $vwl.entry -width 10
   $vwl.entry insert 0 [lindex $gbl(vx,$layer) $vnode]
   label $vwl.unit -text "Km"
   button $vwl.left -text "<<" -command "select_neighbor_vgrid -1 $layer $vnode"
   button $vwl.right -text ">>" -command "select_neighbor_vgrid +1 $layer $vnode"
   pack $vwl.xlabel $vwl.entry $vwl.unit $vwl.left $vwl.right -in $vwl\
       -fill x -side left -pady 5

   # P-Velocity set frame ------------------------------
   frame $vw.set -relief ridge -bd 2
   # Entry
   foreach {title i} {Top vtop Bottom vbot} {
      set f $vw.set.$i
      frame $f -relief groove -bd 2
      label $f.label -text "$title P-Vel." -width 14 -anchor w
      entry $f.entry -width 7
      # Current velocity
      $f.entry insert 0 [lindex $gbl($i,$layer) $vnode]
      label $f.unit -text "Km/sec."
      # Set All button (velocity)
      button $f.setall -text "Set All" -anchor e\
	  -command "set_all_velocity $layer $i $vw.set"
      # pack
      pack $f.label $f.entry $f.unit $f.setall -in $f -side left
      # Standard deviation, in case of ERI-inversion.
      label $f.sdlabel -text "Std.Dev."
      entry $f.sdentry -width 7
      $f.sdentry insert 0 [lindex $gbl($i,SD,$layer) $vnode]
      # Set All button (SD)
      button $f.sdsetall -text "Set All" -anchor e\
	  -command "set_all_sd $layer $i $vw.set"
      if {$gbl(ERI,inv)} {
	 pack $f.sdlabel $f.sdentry $f.sdsetall -in $f -side left
      }
      pack $f -in $vw.set -side top -fill x
   }

   # Vp/Vs (Poisson's ratio) set frame ------------------------------
   set pf $vw.vpvs_frame
   make_vpvs_frame $pf $layer $vnode

   # Key bindings
   bind Entry <Return> "finish_velocity $layer $vnode ok $vw; destroy $vw"

   # Button frame ------------------------------
   frame $vw.button
   # OK 
   button $vw.button.ok -text "OK" -width 10\
       -command "finish_velocity $layer $vnode ok $vw; destroy $vw" 
   # Revert
   button $vw.button.revert -text "Revert" -width 10\
       -command "edit_velocity $XX $YY $name"
   # Cancel
   button $vw.button.cancel -text "Cancel" -width 10\
       -command "finish_velocity $layer $vnode cancel $vw; destroy $vw"
   # Packing of button frame
   pack $vw.button.ok $vw.button.revert $vw.button.cancel\
       -in $vw.button -side left -expand 1
   
   # Packing ------------------------------
   pack $vw.title $vwl $vw.set $pf $vw.button -side top -in $vw \
       -fill x -expand yes
   busy hold .
}
# --------------------------------------------------------------------
# set_all_velocity ---------------------------------------------------
#            Set all the velocity grid in the same layer
# --------------------------------------------------------------------
# [arguments] 
#   layer :  Layer number
#   mode  :  "vtop" or "vbot"
#   w     :  parent widget name
# --------------------------------------------------------------------
proc set_all_velocity {layer mode w} {
    # Global variables --------------------------------------------------
    global gbl
    
    # Main --------------------------------------------------
    # undo information
    set_vgrid_undo_info vgrid.all_velocity $layer -1
    # get velocity value from entry
    set v [$w.$mode.entry get]
    # Update velocity values
    set gbl($mode,$layer) ""
    for {set i 0} {$i < $gbl(n_vnode,$layer)} {incr i} {
	lappend gbl($mode,$layer) $v
    }
}

# --------------------------------------------------------------------
# set_all_sd ---------------------------------------------------------
#            Set all the Velocity SD-value in the same layer
# --------------------------------------------------------------------
# [arguments] 
#   layer :  Layer number
#   mode  :  "vtop" or "vbot"
#   w     :  Parent widget name
# --------------------------------------------------------------------
proc set_all_sd {layer mode w} {
    # Global variables --------------------------------------------------
    global gbl
    
    # Main --------------------------------------------------
    # undo information
    set_vgrid_undo_info vgrid.all_velocity $layer -1
    # get SD from entry
    set sd [$w.$mode.sdentry get]
    # update SD
    set gbl($mode,SD,$layer) ""
    for {set i 0} {$i < $gbl(n_vnode,$layer)} {incr i} {
	lappend gbl($mode,SD,$layer) $sd
    }
}

# --------------------------------------------------------------------
# finish_velocity ----------------------------------------------------
#            Close the velocity edit window
# --------------------------------------------------------------------
# [arguments] 
#   layer :  Layer number
#   vnode :  Grid node number in the layer
#   mode  :  finish mode, "ok" or "cancel"
#   w     :  Parent widget name
# --------------------------------------------------------------------
proc finish_velocity {layer vnode mode w} {
    # Global variables --------------------------------------------------
    global gbl sys wk
    
    # Main --------------------------------------------------
    # Unhighlight the grid --------------------
    $sys(graph) marker delete m($layer,$vnode,v).tmp
    busy release .

    # "ok" mode --------------------
    if ![string compare $mode "ok"] {
	# Get values from entry
	set new(vx) [format "%.3f" [$w.location.entry get]]
	set new(vtop) [format "%.3f" [$w.set.vtop.entry get]]
	set new(vbot) [format "%.3f" [$w.set.vbot.entry get]]
	set new(SD,vtop) [format "%.3f" [$w.set.vtop.sdentry get]]
	set new(SD,vbot) [format "%.3f" [$w.set.vbot.sdentry get]]
	# Vp/Vs
	set pf $w.vpvs_frame
	set new(poi,top) [format "%.4f" [vpvs2poisson [get_float_entry $pf.e1]]]
	set new(SD,poi,top) [format "%.3f" [get_float_entry $pf.e2]]
	set new(poi,bot) [format "%.4f" [vpvs2poisson [get_float_entry $pf.e3]]]
	set new(SD,poi,bot) [format "%.3f" [get_float_entry $pf.e4]]

	if [update_vgrid $layer $vnode $new(vx) $new(vtop) $new(vbot)\
		$new(SD,vtop) $new(SD,vbot) $new(poi,top) $new(poi,bot)\
		$new(SD,poi,top) $new(SD,poi,bot)] {
	    set_vgrid_undo_info vgrid.edit $layer $vnode
	}
	# Grid position ----------
	if {$vnode < 1 || $vnode == [expr $gbl(n_vnode,$layer) - 1]} {
	    # When both end of the structure,
	    bar_message "You can't move edge grids."
	    set new(vx) [lindex $gbl(vx,$layer) $vnode]
	} else {
	    set left [lindex $gbl(vx,$layer) [expr $vnode-1]]
	    set right [lindex $gbl(vx,$layer) [expr $vnode+1]]
	    if {$new(vx) < $left} {
		set new(vx) $left
	    } elseif {$new(vx) > $right} {
		set new(vx) $right
	    }
	}
	set gbl(vx,$layer) [lreplace $gbl(vx,$layer) $vnode $vnode $new(vx)]
	# Velocity and SD ----------
	foreach {i} {vtop vbot} {
	    set gbl($i,$layer) \
		[lreplace $gbl($i,$layer) $vnode $vnode $new($i)]
	    set gbl($i,SD,$layer)\
		[lreplace $gbl($i,SD,$layer) $vnode $vnode  $new(SD,$i)]
	}
	# Poisson's ratio ----------
	switch $gbl(ERI,poisson) {
	   1 {
	      # 1 value per 1 layer
	      set gbl(poi,$layer,top) $new(poi,top)
	      set gbl(poi,SD,$layer,top) $new(SD,poi,top)
	   }
	   2 {
	      # 1 value per 1 grid
	      set gbl(poi,$layer,top)\
		  [lreplace $gbl(poi,$layer,top) $vnode $vnode $new(poi,top)]
	      set gbl(poi,SD,$layer,top)\
		  [lreplace $gbl(poi,SD,$layer,top) $vnode $vnode\
		       $new(SD,poi,top)]
	   }
	   3 {
	      # 2 values per 1 layer
	      foreach {i} {top bot} {
		 set gbl(poi,$layer,$i)\
		     [lreplace $gbl(poi,$layer,$i) $vnode $vnode $new(poi,$i)]
		 set gbl(poi,SD,$layer,$i)\
		     [lreplace $gbl(poi,SD,$layer,$i) $vnode $vnode\
			  $new(SD,poi,$i)]
	      }
	   }
	}
	# Drawing
	draw_grid_layer $layer
	bar_message "Velocity grid was changed in $wk(selected) layer."
    } else {
	bar_message ""
    }
    bind Entry <Return> ""
}

# --------------------------------------------------------------------
# create_grid --------------------------------------------------------
#            Create new velocity grid
# --------------------------------------------------------------------
# [arguments]
#   x  :  x coordinate of mouse cursor. 
#   y  :  y coordinate of mouse cursor. 
# --------------------------------------------------------------------
proc create_grid {x y} {
   # Global variables --------------------------------------------------
   global gbl sys wk
   
   # Main --------------------------------------------------
   # When grid is invisible
   if {$gbl(option,grid) == 0} return
   
   # Get mouse position ------------------------------
   set xz [$sys(graph) invtransform $x $y]
   set x [format "%.3f" [lindex $xz 0]]
   set z [lindex $xz 1]
   
   # inside ?
   if { $x < $gbl(xmin) || $x > $gbl(xmax) } {
      return
   }

   # Highlight the interface
   unselect_interface
   set wk(selected) [search_layer $x $z]
   selection_flash
   
   # Upper and lower side
   if {$wk(selected) >= $gbl(n_interface)} {
      return
   }

   # Insert velocity grid ----------
   # Serach insert postion
   for {set i 0} {$i < $gbl(n_vnode,$wk(selected))} {incr i} {
      if {[lindex $gbl(vx,$wk(selected)) $i] >= $x} {
	 break
      }
   }
   # Compute velocity value
   if {$i > 0} {
      foreach j {vtop vbot} {
	 set $j [expr ([lindex $gbl($j,$wk(selected)) [expr $i -1]]\
			   + [lindex $gbl($j,$wk(selected)) $i])/2.0 ]
	 set SD($j) [expr ([lindex $gbl($j,SD,$wk(selected)) [expr $i -1]]\
			       + [lindex $gbl($j,SD,$wk(selected)) $i])/2.0 ]
      }
   } else {
      foreach j {vtop vbot} {
	 set $j [lindex $gbl($j,$wk(selected)) 0]
	 set SD($j) [lindex $gbl($j,SD,$wk(selected)) 0]
      }
   }
   # Insertion
   foreach {j k} "vx $x vtop $vtop vbot $vbot" {
      set gbl($j,$wk(selected)) [linsert $gbl($j,$wk(selected)) $i $k]
   }
   foreach {j} {vtop vbot} {
      set gbl($j,SD,$wk(selected))\
	  [linsert $gbl($j,SD,$wk(selected)) $i $SD($j)]
   }
   # Poisson's ratio
   if { $gbl(ERI,poisson) >= 2 } {
      if {$i > 0} {
	 foreach j {top bot} {
	    set $j [expr ([lindex $gbl(poi,$wk(selected),$j) [expr $i -1]]\
			      + [lindex $gbl(poi,$wk(selected),$j) $i])/2.0 ]
	    set SD($j) [expr ([lindex $gbl(poi,SD,$wk(selected),$j) [expr $i -1]]\
				  + [lindex $gbl(poi,SD,$wk(selected),$j) $i])/2.0 ]
	 }
      } else {
	 foreach j {top bot} {
	    set $j [lindex $gbl(poi,$wk(selected),$j) 0]
	    set SD($j) [lindex $gbl(poi,SD,$wk(selected),$j) 0]
	 }
      }
      foreach {j k} "top $top bot $bot" {
	 set gbl(poi,$wk(selected),$j)\
	     [linsert $gbl(poi,$wk(selected),$j) $i $k]
      }
      foreach {j} {top bot} {
	 set gbl(poi,SD,$wk(selected),$j)\
	     [linsert $gbl(poi,SD,$wk(selected),$j) $i $SD($j)]
      }
   }

   # undo information
   set_vgrid_undo_info vgrid.create $wk(selected) $i
   # increment
   incr gbl(n_vnode,$wk(selected))
   
   # redraw ----------
   draw_grid_layer $wk(selected)
   bar_message "Added a velocity grid in $wk(selected) layer."
}
# --------------------------------------------------------------------
# delete_velocity ----------------------------------------------------
#            Delete velocity grid
# --------------------------------------------------------------------
# [arguments]
#   name  :  Grid name to delete
# --------------------------------------------------------------------
proc delete_velocity {name} {
    # Global variables --------------------------------------------------
    global gbl sys wk
    
    # Main --------------------------------------------------
    # When invisible grid,
    if {$gbl(option,grid) == 0} return

    # Get layer number and node number. ----------
    set tmp [split [string trim $name "m(,v)"] ","]
    set layer [lindex $tmp 0]
    set vnode [lindex $tmp 1]
    # You cannot delete grids at both end of structure. ----------
    if {$vnode < 1 || $vnode == [expr $gbl(n_vnode,$layer) - 1]} {
	bar_message "You can't delete edge grids."
	# Unhighlight 
	$sys(graph) marker delete $name.tmp
    } else {
	# Confirmation ----------
	set ans [tk_messageBox -title "DELETE: All right?"\
		     -message "Delete this velocity grid. Are you sure?"\
		     -type yesno -icon question]
    }
    # Unhighlight
    $sys(graph) marker delete $name.tmp
    if {$ans=="no"} {
	# Cancel
	bar_message "Delete canceled."
	return
    }
    # Get the v-grid number ----------
    set tmp [split [string trim $name "m(,v)"] ","]
    set vnode [lindex $tmp 1]
    
    # undo information ----------
    set_vgrid_undo_info vgrid.delete $wk(selected) $vnode

    # delete ----------
    # P-velocity
    foreach j {vx vtop vbot} {
	set gbl($j,$wk(selected))\
		[lreplace $gbl($j,$wk(selected)) $vnode $vnode]
    }
    foreach j {vtop vbot} {
	set gbl($j,SD,$wk(selected))\
		[lreplace $gbl($j,SD,$wk(selected)) $vnode $vnode]
    }
    # poisson's ratio
    if { $gbl(ERI,poisson) >= 2 } {
       foreach j {top bot} {
	  set gbl(poi,$wk(selected),$j)\
	      [lreplace $gbl(poi,$wk(selected),$j) $vnode $vnode]
	  set gbl(poi,SD,$wk(selected),$j)\
	      [lreplace $gbl(poi,SD,$wk(selected),$j) $vnode $vnode]
       }
    }

    incr gbl(n_vnode,$wk(selected)) -1

    # redraw ----------
    draw_grid_layer $wk(selected)
    bar_message "Deleted a velocity grid in $wk(selected) layer."
}
# --------------------------------------------------------------------
# move_grid_start ----------------------------------------------------
#            Begin to move velocity grid
# --------------------------------------------------------------------
# [arguments]
#  name   :  Name of the v-grid
#  gx     :  x coordinate of mouse cursor. (Graph widget coordiantes.)
# --------------------------------------------------------------------
proc move_grid_start {name gx} {
    # Global variables --------------------------------------------------
    global gbl sys wk
    
    # Main --------------------------------------------------
    # Get the layer and node number ----------
    set tmp [split [string trim $name "m(,v)"] ","]
    set layer [lindex $tmp 0]
    set vnode [lindex $tmp 1]

    # Cannot move the grid at both end.
    if {$vnode < 1 || $vnode == [expr $gbl(n_vnode,$layer) - 1]} {
	bar_message "You can't move edge grids."
	# Unhighlight
	$sys(graph) marker delete $name.tmp
	return
    }

    # Save the grid information to buffer
    set gbl(move_grid,original_name) $name
    if {$vnode > 0} {
	set gbl(move_grid,left) [lindex $gbl(vx,$layer) [expr $vnode-1]]
    } else {
	set gbl(move_grid,left) $gbl(xmin)
    }
    if {$vnode < [expr $gbl(n_vnode,$layer)-1]} {
	set gbl(move_grid,right) [lindex $gbl(vx,$layer) [expr $vnode+1]]
    } else {
	set gbl(move_grid,right) $gbl(xmax)
    }
    # Display original velocity grid by dotted line.
    $sys(graph) marker configure $gbl(move_grid,original_name).tmp\
	    -outline black -dashes "5 3" -linewidth 1.5

    # Moving grid...
    $sys(graph) marker create line -name moving_grid_name\
	-coords [$sys(graph) marker cget\
		     $gbl(move_grid,original_name) -coords]\
	-outline $wk(col,select) -linewidth 2
    # Key bindigs for moving and finishing
    bind $sys(graph) <B3-Motion>\
	    "move_grid_motion %x %y"
    bind $sys(graph) <ButtonRelease-3>\
	    "move_grid_end %x %y"
    bar_message "Velocity grid ... x = $gx km."
}
# --------------------------------------------------------------------
# move_grid_motion ---------------------------------------------------
#            Grid move, during motion.
# --------------------------------------------------------------------
# [arguments] 
#   x    : x coordinate of mouse cursor. (Graph widget coordiantes.)
#   y    : y coordinate of mouse cursor. (Graph widget coordiantes.)
# --------------------------------------------------------------------
proc move_grid_motion {x y} {
    # Global variables --------------------------------------------------
    global gbl sys wk
    
    # Main --------------------------------------------------
    # When invisible grid,
    if ![string compare $gbl(move_grid,original_name) ""] {
	return;
    }
    set xz [$sys(graph) invtransform $x $y]
    set gx [lindex $xz 0]
    set gz [lindex $xz 1]

    # Cannot over the neighbor grid.
    if {$gx > $gbl(move_grid,right)} {
	set gx $gbl(move_grid,right)
    } elseif {$gx < $gbl(move_grid,left)} {
	set gx $gbl(move_grid,left)
    }
    # Erase moving grid.
    $sys(graph) marker delete moving_grid_name
    
    # Check the mouse postion
    # upper and lower
    set top [interpolate_depth $wk(selected) $gx]
    set bottom [interpolate_depth [expr $wk(selected) + 1] $gx]
    if {[expr ($bottom - $gz)*($gz - $top)] < 0} {
	move_grid_cancel
	return
    }
    # left and right
    if {$gx > $gbl(move_grid,right)} {
	set gx $gbl(move_grid,right)
    } elseif {$gx < $gbl(move_grid,left)} {
	set gx $gbl(move_grid,left)
    }
    
    # Display moving grid.
    $sys(graph) marker create line -name moving_grid_name\
	-coords "$gx $top $gx $bottom" -outline $wk(col,select) -linewidth 2
   
    bar_message [format "Velocity grid ... x = %.3f km." $gx]
}
# --------------------------------------------------------------------
# move_grid_end ------------------------------------------------------
#            Finish moving grid.
# --------------------------------------------------------------------
# [arguments] 
#   x    : x coordinate of mouse cursor. (Graph widget coordiantes.) 
#   y    : y coordinate of mouse cursor. (Graph widget coordiantes.) 
# --------------------------------------------------------------------
proc move_grid_end {x y} {
    # Global variables --------------------------------------------------
    global gbl sys
    
    # Main --------------------------------------------------
    # When no-moving mode,
    if ![string compare $gbl(move_grid,original_name) ""] {
	return;
    }
    # Get layer and node number ----------
    set tmp [split [string trim $gbl(move_grid,original_name) "m(,v)"] ","]
    set layer [lindex $tmp 0]
    set vnode [lindex $tmp 1]

    set xz [$sys(graph) invtransform $x $y]
    set gx [format "%.3f" [lindex $xz 0]]
    set gz [lindex $xz 1]
    # Delete temporary moving grid.
    $sys(graph) marker delete moving_grid_name

    # Check neighbor grid.
    if {$gx > $gbl(move_grid,right)} {
	set gx $gbl(move_grid,right)
    } elseif {$gx < $gbl(move_grid,left)} {
	set gx $gbl(move_grid,left)
    }
   
    # undo information
    set_vgrid_undo_info vgrid.edit $layer $vnode

    # update coordinate
    set gbl(vx,$layer) [lreplace $gbl(vx,$layer) $vnode $vnode $gx]

    # redraw
    bar_message "Velocity grid was set to x = $gx km."
    move_grid_cancel
    draw_grid_layer $layer
}
# --------------------------------------------------------------------
# move_grid_cancel ---------------------------------------------------
#            Cancel moving the grid
# --------------------------------------------------------------------
proc move_grid_cancel {} {
    # Global variables --------------------------------------------------
    global gbl sys
    
    # Main --------------------------------------------------
    # Un-highlight
    $sys(graph) marker delete $gbl(move_grid,original_name).tmp
    set gbl(move_grid,original_name) ""
}
# --------------------------------------------------------------------
# show_velocity ------------------------------------------------------
#            Show velocity value and its gradient at mouse position
# --------------------------------------------------------------------
# [arguments]
#  x  : x coordinate of mouse cursor. (Graph widget coordiantes.)
#  y  : y coordinate of mouse cursor. (Graph widget coordiantes.)
# --------------------------------------------------------------------
proc show_velocity {x y} {
   # Global variables --------------------------------------------------
   global gbl sys
   
   # Main --------------------------------------------------
   # Outside of graph widget 
   if { ![$sys(graph) inside $x $y] \
	    || ![$sys(graph) element exist graph_frame]} {
      return
   }
   # Get mouse position ----------
   set xz [$sys(graph) invtransform $x $y]
   set gx [lindex $xz 0]
   set gz [lindex $xz 1]
   
   # Show velocity ----------
   switch $gbl(option,vel) {
      0 {
	 # no
	 set mes ""
      }
      1 {
	 # P and gradient
	 if { [set vellist [calc_velocity $gx $gz]] == ""} return
	 set vel [lindex $vellist 0]
	 set vvel [lindex $vellist 1]
	 set hvel [lindex $vellist 2]
	 # Output ----------
	 set mes [format\
		      "Vp=%6.3fkm/sec, dVp/dz=%6.3f, dVp/dx=%6.3f at (%8.3fkm, %8.3fkm)"\
		      $vel $vvel $hvel $gx $gz]
      }
      2 {
	 # P and Poisson's ratio
	 if { [set vellist [calc_velocity $gx $gz]] == "" || \
		  [set poilist [calc_poisson $gx $gz]] == "" } return
	 set pvel [lindex $vellist 0]
	 set poi [lindex $poilist 0]
	 set mes [format\
		      "Vp=%6.3fkm/sec, Poisson's ratio=%6.3f at (%8.3fkm, %8.3fkm)"\
		      $pvel $poi $gx $gz]
      }
      3 {
	 # P and S
	 if { [set vellist [calc_velocity $gx $gz]] == "" || \
		  [set poilist [calc_poisson $gx $gz]] == "" } return
	 set pvel [lindex $vellist 0]
	 set svel [s_vel $pvel [lindex $poilist 0]]
	 set mes [format\
		      "Vp=%6.3fkm/sec, Vs=%6.3fkm/sec at (%8.3fkm, %8.3fkm)"\
		      $pvel $svel $gx $gz]
	 
      }
      4 {
	 # S and Poisson's
	 if { [set vellist [calc_velocity $gx $gz]] == "" || \
		  [set poilist [calc_poisson $gx $gz]] == "" } return
	 set pvel [lindex $vellist 0]
	 set poi [lindex $poilist 0]
	 set svel [s_vel $pvel $poi]
	 set mes [format\
		      "Vs=%6.3fkm/sec, Poisson's ratio=%6.3f, at (%8.3fkm, %8.3fkm)"\
		      $svel $poi $gx $gz]
      }
      5 {
	 # P and S velocity and Poisson's
	 if { [set vellist [calc_velocity $gx $gz]] == "" || \
		  [set poilist [calc_poisson $gx $gz]] == "" } return
	 set pvel [lindex $vellist 0]
	 set poi [lindex $poilist 0]
	 set svel [s_vel $pvel $poi]
	 set mes [format\
		      "Vp=%6.3fkm/sec, Vs=%6.3fkm/sec. Pois's ratio=%6.3f, at (%8.3fkm, %8.3fkm)"\
		      $pvel $svel $poi $gx $gz]
      }
   }
   
   # Show ----------
   bar_message $mes
}
# --------------------------------------------------------------------
# calc_velocity ------------------------------------------------------
#            Compute velocity and its gradient at the point, P.
# --------------------------------------------------------------------
# [arguments]
#  x  : x-coordinate of P
#  z  : y-coordinate of P
# --------------------------------------------------------------------
proc calc_velocity {x z} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set info [find_trape $x $z]
   if {$info == ""} {
      return ""
   } else {
      set i [lindex $info 0]
      set x1 [lindex $info 1]
      set x2 [lindex $info 2]
   }

   # Compute velocity at the vertexes
   set v1 [calc_v_layer $i $x1]
   set v2 [calc_v_layer $i $x2]
   set v1top [lindex $v1 0]
   set v2top [lindex $v2 0]
   set v1bot [lindex $v1 1]
   set v2bot [lindex $v2 1]

   # Retrun
   return [zelt_velocity $i $x1 $v1top $v1bot $x2 $v2top $v2bot $x $z]
}

# --------------------------------------------------------------------
# find_trape ---------------------------------------------------------
#            find trapezoid of the neighbor of point P.
# --------------------------------------------------------------------
# [arguments]
#  x  : x-coordinate of P
#  z  : y-coordinate of P
# --------------------------------------------------------------------
proc find_trape {x z} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set i [find_layer $x $z]
   if {$i == ""} return

   # Searching the trapezoid which contains P --------------------
   # Divide the $i-th layer into trapezoids
   set xlist [lsort -real "$gbl(x,$i) $gbl(x,[expr $i+1]) $gbl(vx,$i)"]
   # compute the vertex of the trapezoids
   set j 1
   while {1} {
      set x1 [lindex $xlist [expr $j-1]]
      set x2 [lindex $xlist $j]
      if {$x1 <= $x && $x2 >= $x && [expr $x1 - $x2] != 0.0} {
	 break;
      }
      incr j
      if {$j >= [llength $xlist]} {
	 bar_message "ERROR: I'm afraid something wrong with the model."
	 return
      }
   }
   # return trapezoid information, <layer> <left> <right>
   return "$i $x1 $x2"
}

# --------------------------------------------------------------------
# find_layer ---------------------------------------------------------
#            find layer which contain point P.
# --------------------------------------------------------------------
# [arguments]
#  x  : x-coordinate of P
#  z  : y-coordinate of P
# --------------------------------------------------------------------
proc find_layer {x z} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   # P is inside model?
   if {$x < $gbl(xmin) || $x > $gbl(xmax)} {
      return ""
   }
   # The layer number in which P is --------------------
   set i 2
   while {$i <= $gbl(n_interface)} {
      if { [interpolate_depth $i $x] >= $z} {
	 break;
      }
      incr i
   }
   # P is in model? (Vertically)
   if {$i > $gbl(n_interface) || $z < [interpolate_depth [expr $i-1] $x]} {
      return ""
   }
   return [expr $i - 1]
}

# --------------------------------------------------------------------
# calc_v_layer -------------------------------------------------------
#            Compute velocity at the interface of $i-th layer at $x.
# --------------------------------------------------------------------
# [arguments]
#  i  :  layer number.
#  x  :  x-coordinate.
# --------------------------------------------------------------------
proc calc_v_layer {i x} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set j2 [search_vgrid $i $x]
   set j1 [expr $j2 - 1]
   set x1 [lindex $gbl(vx,$i) $j1]
   set x2 [lindex $gbl(vx,$i) $j2]
   foreach mode {vtop vbot} {
      set v1 [lindex $gbl($mode,$i) $j1]
      set v2 [lindex $gbl($mode,$i) $j2]
      set $mode [expr ($v1*($x2 - $x) + $v2*($x - $x1)) / ($x2 - $x1)]
   }
   # return
   return "$vtop $vbot"
}

# --------------------------------------------------------------------
# calc_SD_layer ------------------------------------------------------
#         Compute SD of velocity at the interface of $i-th layer at $x.
# --------------------------------------------------------------------
# [arguments]
#  i  :  layer number.
#  x  :  x-coordinate.
# --------------------------------------------------------------------
proc calc_SD_layer {i x} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   set j2 [search_vgrid $i $x]
   set j1 [expr $j2 - 1]
   set x1 [lindex $gbl(vx,$i) $j1]
   set x2 [lindex $gbl(vx,$i) $j2]
   foreach mode {vtop vbot} {
      set v1 [lindex $gbl($mode,SD,$i) $j1]
      set v2 [lindex $gbl($mode,SD,$i) $j2]
      set $mode [expr ($v1*($x2 - $x) + $v2*($x - $x1)) / ($x2 - $x1)]
   }
   # return
   return "$vtop $vbot"
}

# --------------------------------------------------------------------
# search_vgrid -------------------------------------------------------
#            Search nearest vgrid (right side)
# --------------------------------------------------------------------
# [arguments]
#  i  :  layer number.
#  x  :  x-coordinate.
# --------------------------------------------------------------------
proc search_vgrid {i x} {
   # Global variables --------------------------------------------------
   global gbl
   
   # Main --------------------------------------------------
   # Search the velocity grids of both end of $x
   set j2 1
   while {$j2 < $gbl(n_vnode,$i)} {
      set x1 [lindex $gbl(vx,$i) [expr $j2 -1]]
      set x2 [lindex $gbl(vx,$i) $j2]
      if { $x1 <= $x && $x <= $x2} {
	 break
      }
      incr j2
   }
   return $j2
}
# --------------------------------------------------------------------
# zelt_velocity ------------------------------------------------------
#            Interpolate velocity in the trapezoide at (x,z)
#                (liner interpolation, the same algorithm as Zelt.)
# --------------------------------------------------------------------
# [arguments]
#  i  : Layer number
#  x1 : x-coordinate of left side of trapezoid.
#  v1 : velocity of upper-left vertex.
#  v3 : velocity of lower-left vertex.
#  x2 : x-coordinate of right side of trapezoid.
#  v2 : velocity of upper-right vertex.
#  v4 : velocity of lower-right vertex.
#  x  : x-coordinate.
#  z  : z-coordinate.
# --------------------------------------------------------------------
proc zelt_velocity {i x1 v1 v3 x2 v2 v4 x z} {
    # Global variables --------------------------------------------------
    global gbl
    
    # Main --------------------------------------------------
    # Compute the slopes and intercept of the upper and lower side of trapezoid
    set z1t [interpolate_depth $i $x1]
    set z1b [interpolate_depth [expr $i+1] $x1]
    set z2t [interpolate_depth $i $x2]
    set z2b [interpolate_depth [expr $i+1] $x2]
    set s1 [expr ($z2t - $z1t)/($x2 - $x1)]
    set b1 [expr $z1t - $s1*$x1]
    set s2 [expr ($z2b - $z1b)/($x2 - $x1)]
    set b2 [expr $z1b - $s2*$x1]
    # compute coefficients
    set c1 [expr $s2*($x2*$v1 - $x1*$v2) + $b2*($v2 - $v1)\
	    - $s1*($x2*$v3 - $x1*$v4) - $b1*($v4 - $v3)]
    set c2 [expr $s2*($v2 - $v1) - $s1*($v4 - $v3)]
    set c3 [expr $x1*$v2 - $x2*$v1 + $x2*$v3 - $x1*$v4]
    set c4 [expr $v1 - $v2 + $v4 - $v3]
    set c5 [expr $b2*($x2*$v1 - $x1*$v2) - $b1*($x2*$v3 - $x1*$v4)]
    set c6 [expr ($s2 - $s1)*($x2 - $x1)]
    set c7 [expr ($b2 - $b1)*($x2 - $x1)]
    # Compute velocity and its gradients
    set bunshi [expr ($c1*$x + $c2*$x*$x + $c3*$z + $c4*$x*$z + $c5)]
    set bunbo [expr ($c6*$x + $c7)]
    if { $bunbo == 0.0 } { return "" }
    set vel [expr $bunshi/$bunbo]
    set vvel [expr ($c3 + $c4*$x)/$bunbo]
    set hvel [expr ($bunbo*($c1 + $c2*$x +$c4*$z) - $c6*$bunshi)\
	    /($bunbo*$bunbo)]
    return "$vel $vvel $hvel"
}
# --------------------------------------------------------------------
# search_layer -------------------------------------------------------
#        Determine layer number in which mouse pointer is
# --------------------------------------------------------------------
# [arguments]
#   x  : x coordinate of mouse cursor. (Graph widget coordiantes.)
#   z  : y coordinate of mouse cursor. (Graph widget coordiantes.)
# --------------------------------------------------------------------
proc search_layer {x z} {
    # Global variables --------------------------------------------------
    global gbl
    
    # Main --------------------------------------------------
    set i 2
    while {$i <= $gbl(n_interface)} {
	if { [interpolate_depth $i $x] >= $z} {
	    break;
	}
	incr i
    }
    incr i -1
    return $i
}
# --------------------------------------------------------------------
# set_vgrid_undo_info ------------------------------------------------
#        Update undo information, when you edit the velocity grid.
# --------------------------------------------------------------------
# [arguments]
#   m  : mode(vgrid.edit, vgrid.create, vgrid.delete, vgrid.all_velocity)
#   l  : Layer number
#   i  : Grid node number
# --------------------------------------------------------------------
proc set_vgrid_undo_info {m l i} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # Common undo information
   set n $wk(undo,n)
   set wk(undo,index,$n) $i
   set wk(undo,mode,$n) $m
   if {$i >= 0} {
      # Only one velocity grid
      set wk(undo,x,$n) [lindex $gbl(vx,$l) $i]
      set wk(undo,z,$n) dum
      set wk(undo,z,SD,$n) dum
      set wk(undo,vtop,$n) [lindex $gbl(vtop,$l) $i]
      set wk(undo,vbot,$n) [lindex $gbl(vbot,$l) $i]
      set wk(undo,vtop,SD,$n) [lindex $gbl(vtop,SD,$l) $i]
      set wk(undo,vbot,SD,$n) [lindex $gbl(vbot,SD,$l) $i]
      set wk(undo,interface,$n) $l
      if {$gbl(ERI,poisson) <= 1} { set i 0 }
      set wk(undo,poi,top,$n) [lindex $gbl(poi,$l,top) $i]
      set wk(undo,poi,bot,$n) [lindex $gbl(poi,$l,bot) $i]
      set wk(undo,poi,SD,top,$n) [lindex $gbl(poi,SD,$l,top) $i]
      set wk(undo,poi,SD,bot,$n) [lindex $gbl(poi,SD,$l,bot) $i]
   } elseif { $i == -1 }  {
      # All the velocity grid in the same layer
      set wk(undo,x,$n) dum
      set wk(undo,z,$n) dum
      set wk(undo,z,SD,$n) dum
      set wk(undo,vtop,$n) "$gbl(vtop,$l)"
      set wk(undo,vbot,$n) "$gbl(vbot,$l)"
      set wk(undo,vtop,SD,$n) "$gbl(vtop,SD,$l)"
      set wk(undo,vbot,SD,$n) "$gbl(vbot,SD,$l)"
      set wk(undo,interface,$n) $l
      set wk(undo,poi,top,$n) "$gbl(poi,$l,top)"
      set wk(undo,poi,bot,$n) "$gbl(poi,$l,bot)"
      set wk(undo,poi,SD,top,$n) "$gbl(poi,SD,$l,top)"
      set wk(undo,poi,SD,bot,$n) "$gbl(poi,SD,$l,bot)"
   } else {
      # All layer
      set wk(undo,x,$n) "$gbl(vx,$l)"
      set wk(undo,z,$n) dum
      set wk(undo,z,SD,$n) dum
      set wk(undo,vtop,$n) "$gbl(vtop,$l)"
      set wk(undo,vbot,$n) "$gbl(vbot,$l)"
      set wk(undo,vtop,SD,$n) "$gbl(vtop,SD,$l)"
      set wk(undo,vbot,SD,$n) "$gbl(vbot,SD,$l)"
      set wk(undo,interface,$n) $l
      set wk(undo,poi,top,$n) "$gbl(poi,$l,top)"
      set wk(undo,poi,bot,$n) "$gbl(poi,$l,bot)"
      set wk(undo,poi,SD,top,$n) "$gbl(poi,SD,$l,top)"
      set wk(undo,poi,SD,bot,$n) "$gbl(poi,SD,$l,bot)"
   }
    incr wk(undo,n)
    undo_check
}
# --------------------------------------------------------------------
# highlight_grid ------------------------------------------------
#        Highlight the grid for a while
# --------------------------------------------------------------------
# [arguments]
#   l  : Layer number
#   i  : Grid node number
# --------------------------------------------------------------------
proc highlight_grid {l i} {
    # Global variables --------------------------------------------------
    global gbl sys wk

    # Main --------------------------------------------------
    set x [lindex $gbl(vx,$l) $i]
    set z1 [interpolate_depth $l $x]
    set z2 [interpolate_depth [expr $l+1] $x]
    $sys(graph) marker create line -name tmp_highlight($l,$i) \
	    -coords "$x $z1 $x $z2" -linewidth 2 -outline $wk(col,select)
    eval after 1000 {
	eval $sys(graph) marker delete\
		[$sys(graph) marker names tmp_highlight($l,*)]
    }
}
# --------------------------------------------------------------------
# update_vgrid -------------------------------------------------------
#        Status check of v-grid, if edited, return 1.
# --------------------------------------------------------------------
# [arguments]
#   l  : layer bumber
#   i  : grid node number
#   x  : New x-coordinate of grid node.
#   t  : New upper velocity
#   b  : New lower velocity
#   st : New upper SD
#   sb : New lower SD
#   pt : New poisson's ratio top
#   pb : New poisson's ratio bottom
#   pst: New SD of poisson's ratio top
#   psb: New SD of poisson's ratio bot
# --------------------------------------------------------------------
proc update_vgrid {l i x t b st sb pt pb pst psb} {
    # Global variables --------------------------------------------------
    global gbl

    # Main --------------------------------------------------
    # Check p-velocity and its SD
    if {    $x  != [lindex $gbl(vx,$l) $i]      || \
	    $t  != [lindex $gbl(vtop,$l) $i]    || \
	    $b  != [lindex $gbl(vbot,$l) $i]    || \
	    $st != [lindex $gbl(vtop,SD,$l) $i] || \
	    $sb != [lindex $gbl(vbot,SD,$l) $i] } {
	return 1
    }
    # Check poisson's ratio relation
    switch $gbl(ERI,poisson) {
       0 {
	  # none
	  return 0
       }
       1 {
	  # 1 value per 1 layer
	  set cpt $gbl(poi,$l,top)
	  set cpb $pb
	  set cpst $gbl(poi,SD,$l,top)
	  set cpsb $psb
       }
       2 {
	  # 1 value per 1 grid
	  set cpt [lindex $gbl(poi,$l,top) $i]
	  set cpb ""
	  set cpst [lindex $gbl(poi,SD,$l,top) $i]
	  set cpsb ""
       }
       3 {
	  # 2 values per 1 layer
	  set cpt [lindex $gbl(poi,$l,top) $i]
	  set cpb [lindex $gbl(poi,$l,bot) $i]
	  set cpst [lindex $gbl(poi,SD,$l,top) $i]
	  set cpsb [lindex $gbl(poi,SD,$l,bot) $i]
       }
       default {
	  # the others (cannot exist...)
	  return 0
       }
    }
    if {$pt != $cpt || $pb != $cpb || $pst != $cpst || $psb != $cpsb } {
       return 1
    }
    return 0
}

# --------------------------------------------------------------------
# get_float_entry ----------------------------------------------------
#        Get float %.3f value fron the specified entry widget
# --------------------------------------------------------------------
# [arguments]
#   w   : entry widget path
# --------------------------------------------------------------------
proc get_float_entry {w} {
   set v [$w get]
   if {$v == ""} {return 0.0}
   return [format "%.3f" $v]
}

# --------------------------------------------------------------------
# vgrid_regular_spacing ----------------------------------------------
#        Rearrange velocity grids to be at regular spacing
# --------------------------------------------------------------------
proc vgrid_regular_spacing {} {
   # Global variables --------------------------------------------------
   global gbl wk

   # Main --------------------------------------------------
   if {$wk(selected) == $gbl(n_interface)} {
      bar_message "ERROR: Select a layer first."
      return
   }

   # Invoke window to enter a spacing --------------------
   set w .vgrid_spacing
   catch {destroy $w}
   toplevel $w
   wm title $w "Set a regular spacing for v-grid in $wk(selected) layer."

   # Title frame ------------------------------
   title_frame $w.title "Regular spacing : Layer $wk(selected)"

   # Parameter frame ------------------------------
   set pf [frame $w.parameter_frame]
   set wk(spacing) 10.0
   make_entry $pf spacing "Regular spacing" wk(spacing) km

   # Botton frame ------------------------------
   set bf [button_frame $w $w.button_frame \
	       "apply" "vgrid_apply_regular_spacing"]
   
   # Packing ------------------------------
   pack $w.title $pf $bf -side top -fill x
   # busy hold .
   grab $w
}

# --------------------------------------------------------------------
# vgrid_apply_regular_spacing ----------------------------------------
#        Rearrange velocity grids to be at regular spacing
# --------------------------------------------------------------------
proc vgrid_apply_regular_spacing {} {
   # Global variables --------------------------------------------------
   global gbl wk

   # Main --------------------------------------------------
   set i $wk(selected)
   set s [expr $wk(spacing) * 1.0]
   if {$s <= 0.0} {
      tk_messageBox -type ok -icon error -message \
	  "Invalid spacing."
      return
   }

   # undo information
   set_vgrid_undo_info vgrid.all_velocity $i -1

   # make new layer data
   set x $gbl(xmin)
   set vx {}
   set vtop {}
   set vbot {}
   set vtopSD {}
   set vbotSD {}
   while {$x < $gbl(xmax)} {
      # vx
      lappend vx $x
      # vtop and vbot
      set j2 1
      while {1} {
	 set x2 [lindex $gbl(vx,$i) $j2]
	 if {$x2 >= $x} {
	    break;
	 }
	 incr j2
      }
      set j1 [expr $j2 - 1]
      set x1 [lindex $gbl(vx,$i) $j1]
      lappend vtop \
	  [expr ( [lindex $gbl(vtop,$i) $j1] * ($x2 - $x) \
		      + [lindex $gbl(vtop,$i) $j2] * ($x - $x1) ) / ($x2 - $x1)]
      lappend vbot \
	  [expr ( [lindex $gbl(vbot,$i) $j1] * ($x2 - $x) \
		      + [lindex $gbl(vbot,$i) $j2] * ($x - $x1) ) / ($x2 - $x1)]
      lappend vtopSD \
	  [expr ( [lindex $gbl(vtop,SD,$i) $j1] * ($x2 - $x) \
		      + [lindex $gbl(vtop,SD,$i) $j2] * ($x - $x1) ) \
	       / ($x2 - $x1)]
      lappend vbotSD \
	  [expr ( [lindex $gbl(vbot,SD,$i) $j1] * ($x2 - $x) \
		      + [lindex $gbl(vbot,SD,$i) $j2] * ($x - $x1) ) \
	       / ($x2 - $x1)]
      set x [expr $x + $s]
   }

   # right end
   lappend vx $gbl(xmax)
   lappend vtop [lindex $gbl(vtop,$i) end]
   lappend vbot [lindex $gbl(vbot,$i) end]
   lappend vtopSD [lindex $gbl(vtop,SD,$i) end]
   lappend vbotSD [lindex $gbl(vbot,SD,$i) end]
   
   # poisson's ratio
   if {$gbl(ERI,poisson) == 2} {
      # 1 value per 1 grid
      set poi [lindex $gbl(poi,$i,top) 0]
      set poiSD [lindex $gbl(poi,SD,$i,top) 0]
      set gbl(poi,$i,top) {}
      set gbl(poi,SD,$i,top) {}
      for {set j 0} {$j < [llength $vx]} {incr j} {
	 lappend gbl(poi,$i,top) $poi
	 lappend gbl(poi,SD,$i,top) $poiSD
      }
   }

   # update
   set gbl(vx,$i) $vx
   set gbl(vtop,$i) $vtop
   set gbl(vbot,$i) $vbot
   set gbl(vtop,SD,$i) $vtopSD
   set gbl(vbot,SD,$i) $vbotSD
   set gbl(n_vnode,$i) [llength $vx]

   draw_grid_layer $i
}

# --------------------------------------------------------------------
# highlight_multiple_vgrid -------------------------------------------
#        highlight multiple v-grid
# --------------------------------------------------------------------
proc highlight_multiple_vgrid {val mode layer} {
   # Global variables --------------------------------------------------
   global gbl sys wk
   
   # Main --------------------------------------------------
   # un-highlight
   eval $sys(graph) marker delete \
       [$sys(graph) marker names tmp.highlight.vgrid(*)]

   # check
   set vx(min) $wk(vxmin)
   set vx(max) $wk(vxmax)
   set vx($mode) $val
   regexp {[\-\d\.]*} $vx(min) left
   regexp {[\-\d\.]*} $vx(max) right
   if { [string length $left] < 1 || [string length $right] < 1 || 
	$left > $right } {
      return 1
   }
   set n [llength $gbl(vx,$layer)]
   for {set i 0} {$i < $n} {incr i} {
      set x [lindex $gbl(vx,$layer) $i]
      if {$x > $right} {
	 break
      }
      if {$x >= $left} {
	 set z1 [interpolate_depth $layer $x]
	 set z2 [interpolate_depth [expr $layer+1] $x]
	 $sys(graph) marker create line -name tmp.highlight.vgrid($i) \
	     -coords "$x $z1 $x $z2" -linewidth 2 -outline $wk(col,select)
      }
   }
   return 1
}

# --------------------------------------------------------------------
# edit_multiple_vgrid ------------------------------------------------
#        Edit multiple v-grid simultaneously
# --------------------------------------------------------------------
proc edit_multiple_vgrids {} {
   # Global variables --------------------------------------------------
   global gbl wk sys

   # Main --------------------------------------------------
   if {$wk(selected) == $gbl(n_interface)} {
      bar_message "ERROR: Select a layer first."
      return
   }

   # Invoke window to enter a spacing --------------------
   set w .vgrid_simul
   catch {destroy $w}
   toplevel $w
   wm title $w "Edit multiple v-grids in layer $wk(selected)"

   # Title frame ------------------------------
   title_frame $w.title "Edit multiple v-grids in layer $wk(selected)"

   # x-range frame ------------------------------
   set xf [frame $w.xrange_frame -relief groove -bd 2]
   set wk(vxmin) $gbl(xmin)
   set wk(vxmax) $gbl(xmax)
   label $xf.label1 -text "X-range  from:"
   entry $xf.from -width 7 -textvar wk(vxmin) -validate key \
       -vcmd "highlight_multiple_vgrid %P min $wk(selected)" 
   label $xf.label2 -text "(km)    to:"
   entry $xf.to -width 7 -textvar wk(vxmax) -validate key \
       -vcmd "highlight_multiple_vgrid %P max $wk(selected)"
   label $xf.unit -text "(km)"
   grid $xf.label1 $xf.from $xf.label2 $xf.to $xf.unit -sticky news

   # P-velocity frame ------------------------------
   set pf [frame $w.vel]
   foreach {title i} {Top vtop Bottom vbot} {
      set f $pf.$i
      frame $f -relief groove -bd 2
      label $f.label -text "$title P-Vel." -width 14 -anchor w
      entry $f.entry -width 7
      # initial velocity
      $f.entry insert 0 [lindex $gbl($i,$wk(selected)) 0]
      label $f.unit -text "Km/sec."
      # Set All button (velocity)
      button $f.setall -text "apply" -anchor e\
	  -command "set_multiple_velocity $wk(selected) $i $pf"
      # pack
      pack $f.label $f.entry $f.unit $f.setall -in $f -side left

      # Standard deviation
      label $f.sdlabel -text "Std.Dev."
      entry $f.sdentry -width 7
      $f.sdentry insert 0 [lindex $gbl($i,SD,$wk(selected)) 0]
      # Set All button (SD)
      button $f.sdsetall -text "apply" -anchor e\
	  -command "set_multiple_sd $wk(selected) $i $pf"
      if {$gbl(ERI,inv)} {
	 pack $f.sdlabel $f.sdentry $f.sdsetall -in $f -side left
      }
      pack $f -in $pf -side top -fill x
   }

   # Vp/Vs (Poisson's ratio) set frame ------------------------------
   set vf [frame $w.vpvs_frame -relief groove -bd 2]
   switch $gbl(ERI,poisson) {
      0 -
      1 {
	 # none
	 # 1 value per 1 layer
      }
      2 {
	 # 1 value per 1 grid
	 label $vf.l1 -text "Vp/Vs of grids"
	 entry $vf.e1 -width 7
	 label $vf.poi1 -text ""
	 button $vf.b1 -text "apply" \
	     -command "set_multiple_poisson $vf.e1 $wk(selected) top"
	 set wl {$vf.l1 $vf.e1 $vf.poi1 $vf.b1}
	 set vpvs [poisson2vpvs [lindex $gbl(poi,$wk(selected),top) 0]]
	 $vf.e1 insert 0 $vpvs
	 $vf.e1 conf -validate key -vcmd "show_pois_label %P $vf.poi1"
	 show_pois_label $vpvs $vf.poi1
	 if {$gbl(ERI,inv)} {
	    label $vf.l2 -text "   Std.Dev."
	    entry $vf.e2 -width 7
	    button $vf.b2 -text "apply"\
		-command "set_multiple_poisson_SD $vf.e2 $wk(selected) top"
	    set wl [concat $wl {$vf.l2 $vf.e2 $vf.b2}]
	    $vf.e2 insert 0 [lindex $gbl(poi,SD,$wk(selected),top) 0]
	 }
	 eval grid $wl -sticky news
      }
      3 {
	 # 2 values per 1 grid
	 foreach {k i} {"Top" top "Bottom" bot } {
	    label $vf.l1($i) -text "Vp/Vs of grids"
	    entry $vf.e1($i) -width 7
	    label $vf.poi1($i) -text ""
	    button $vf.b1($i) -text "apply" \
		-command "set_multiple_poisson $vf.e1($i) $wk(selected) $i"
	    set wl {$vf.l1($i) $vf.e1($i) $vf.poi1($i) $vf.b1($i)}
	    set vpvs [poisson2vpvs [lindex $gbl(poi,$wk(selected),$i) 0]]
	    $vf.e1($i) insert 0 $vpvs
	    $vf.e1($i) conf -validate key -vcmd "show_pois_label %P $vf.poi1($i)"
	    show_pois_label $vpvs $vf.poi1($i)
	    if {$gbl(ERI,inv)} {
	       label $vf.l2($i) -text "   Std.Dev."
	       entry $vf.e2($i) -width 7
	       button $vf.b2($i) -text "apply"\
		   -command "set_multiple_poisson_SD $vf.e2($i) $wk(selected) $i"
	       set wl [concat $wl {$vf.l2($i) $vf.e2($i) $vf.b2($i)}]
	       $vf.e2($i) insert 0 [lindex $gbl(poi,SD,$wk(selected),$i) 0]
	    }
	    eval grid $wl -sticky news

	 }
      }
   }

   # Botton frame ------------------------------
   set bf [button_frame $w $w.button_frame]
   
   # Packing ------------------------------
   pack $w.title $xf $pf $vf $bf -side top -fill x
   # busy hold .
   grab $w

   # un-highlight
   tkwait window $w
   eval $sys(graph) marker delete \
       [$sys(graph) marker names tmp.highlight.vgrid(*)]
}

# --------------------------------------------------------------------
# set_multiple_velocity ----------------------------------------------
#            Set multiple velocity, between $wk(vxmin) and $wk(vxmax)
# --------------------------------------------------------------------
#   layer :  Layer number
#   mode  :  "vtop" or "vbot"
#   w     :  parent widget name
# --------------------------------------------------------------------
proc set_multiple_velocity {layer mode w} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # undo information
   set_vgrid_undo_info vgrid.all_velocity $layer -1
   # get velocity value from entry
   set v [$w.$mode.entry get]
   # Update velocity values
   set prev $gbl($mode,$layer)
   set gbl($mode,$layer) ""
   for {set i 0} {$i < $gbl(n_vnode,$layer)} {incr i} {
      set x [lindex $gbl(vx,$layer) $i]
      if { $x >= $wk(vxmin) && $x <= $wk(vxmax) } {
	 lappend gbl($mode,$layer) $v
      } else {
	 lappend gbl($mode,$layer) [lindex $prev $i]
      }
   }
}

# --------------------------------------------------------------------
# set_multiple_sd ----------------------------------------------------
#            Set multiple SD of velocity of a layer in x-range
# --------------------------------------------------------------------
#   layer :  Layer number
#   mode  :  "vtop" or "vbot"
#   w     :  Parent widget name
# --------------------------------------------------------------------
proc set_multiple_sd {layer mode w} {
   # Global variables --------------------------------------------------
   global gbl wk
   
   # Main --------------------------------------------------
   # undo information
   set_vgrid_undo_info vgrid.all_velocity $layer -1
   # get SD from entry
   set sd [$w.$mode.sdentry get]
   # update SD
   set prev $gbl($mode,SD,$layer)
   set gbl($mode,SD,$layer) ""
   for {set i 0} {$i < $gbl(n_vnode,$layer)} {incr i} {
      set x [lindex $gbl(vx,$layer) $i]
      if { $x >= $wk(vxmin) && $x <= $wk(vxmax) } {
	 lappend gbl($mode,SD,$layer) $sd
      } else {
	 lappend gbl($mode,SD,$layer) [lindex $prev $i]
      }
   }
}

